from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Dict, Iterable, Sequence, Tuple

from dbt_semantic_interfaces.implementations.time_spine import PydanticTimeSpineCustomGranularityColumn
from dbt_semantic_interfaces.protocols import SemanticManifest
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow_semantics.collection_helpers.lru_cache import typed_lru_cache
from metricflow_semantics.errors.error_classes import SemanticManifestConfigurationError
from metricflow_semantics.specs.time_dimension_spec import DEFAULT_TIME_GRANULARITY, TimeDimensionSpec
from metricflow_semantics.sql.sql_table import SqlTable
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class TimeSpineSource:
    """A calendar table. Should contain at least one column with dates/times that map to a standard granularity.

    Dates should be contiguous. May also contain custom granularity columns.
    """

    sql_table: SqlTable
    # Name of the column in the table that contains date/time values that map to a standard granularity.
    base_column: str = "ds"
    # The time granularity of the base column.
    base_granularity: TimeGranularity = DEFAULT_TIME_GRANULARITY
    custom_granularities: Sequence[PydanticTimeSpineCustomGranularityColumn] = ()

    @staticmethod
    def build_standard_time_spine_sources(
        semantic_manifest: SemanticManifest,
    ) -> Dict[TimeGranularity, TimeSpineSource]:
        """Creates a time spine source based on what's in the manifest."""
        time_spine_sources = {
            time_spine.primary_column.time_granularity: TimeSpineSource(
                sql_table=SqlTable.from_string(time_spine.node_relation.relation_name),
                base_column=time_spine.primary_column.name,
                base_granularity=time_spine.primary_column.time_granularity,
                custom_granularities=tuple(
                    [
                        PydanticTimeSpineCustomGranularityColumn(
                            name=custom_granularity.name, column_name=custom_granularity.column_name
                        )
                        for custom_granularity in time_spine.custom_granularities
                    ]
                ),
            )
            for time_spine in semantic_manifest.project_configuration.time_spines
        }

        # For backward compatibility: if legacy time spine config exists in the manifest, add that time spine here for
        # backward compatibility. Ignore it if there is a new time spine config with the same granularity.
        legacy_time_spines = semantic_manifest.project_configuration.time_spine_table_configurations
        for legacy_time_spine in legacy_time_spines:
            if legacy_time_spine.grain not in time_spine_sources:
                time_spine_sources[legacy_time_spine.grain] = TimeSpineSource(
                    sql_table=SqlTable.from_string(legacy_time_spine.location),
                    base_column=legacy_time_spine.column_name,
                    base_granularity=legacy_time_spine.grain,
                )

        # Sanity check: this should have been validated during manifest parsing.
        if not time_spine_sources:
            raise SemanticManifestConfigurationError(
                "At least one time spine must be configured to use the semantic layer, but none were found."
            )

        return time_spine_sources

    @staticmethod
    @typed_lru_cache
    def build_custom_time_spine_sources(time_spine_sources: Sequence[TimeSpineSource]) -> Dict[str, TimeSpineSource]:
        """Creates a set of time spine sources with custom granularities based on what's in the manifest."""
        return {
            custom_granularity.name: time_spine_source
            for time_spine_source in time_spine_sources
            for custom_granularity in time_spine_source.custom_granularities
        }

    @staticmethod
    def build_custom_granularities(time_spine_sources: Iterable[TimeSpineSource]) -> Dict[str, ExpandedTimeGranularity]:
        """Creates a set of supported custom granularities based on what's in the manifest."""
        return {
            custom_granularity.name: ExpandedTimeGranularity(
                name=custom_granularity.name, base_granularity=time_spine_source.base_granularity
            )
            for time_spine_source in time_spine_sources
            for custom_granularity in time_spine_source.custom_granularities
        }

    @staticmethod
    def choose_time_spine_sources(
        required_time_spine_specs: Sequence[TimeDimensionSpec],
        time_spine_sources: Dict[TimeGranularity, TimeSpineSource],
    ) -> Tuple[TimeSpineSource, ...]:
        """Determine which time spine sources to use to satisfy the given specs.

        Custom grains can only use the time spine where they are defined. For standard grains, this will choose the time
        spine with the largest granularity that is compatible with all required standard grains. This ensures max efficiency
        for the query by minimizing the number of time spine joins and the amount of aggregation required. Example:
        - Time spines available: SECOND, MINUTE, DAY
        - Time granularities needed for request: HOUR, DAY
        --> Selected time spine: MINUTE
        """
        assert required_time_spine_specs, (
            "Choosing time spine source requires time spine specs, but the `required_time_spine_specs` param is empty. "
            "This indicates internal misconfiguration."
        )

        # Each custom grain can only be satisfied by one time spine.
        custom_time_spines = TimeSpineSource.build_custom_time_spine_sources(tuple(time_spine_sources.values()))
        required_time_spines = {
            custom_time_spines[spec.time_granularity.name]
            for spec in required_time_spine_specs
            if spec.time_granularity and spec.has_custom_grain
        }

        # Standard grains can be satisfied by any time spine with a base grain that's <= the standard grain. If date part was
        # requested instead of grain, assume DAY since is is the largest grain that's compatible with all supported date parts.
        smallest_required_standard_grain = min(
            TimeGranularity.DAY if spec.time_granularity is None else spec.time_granularity.base_granularity
            for spec in required_time_spine_specs
        )
        compatible_time_spines_for_standard_grains = {
            grain: time_spine_source
            for grain, time_spine_source in time_spine_sources.items()
            if grain.to_int() <= smallest_required_standard_grain.to_int()
        }
        if len(compatible_time_spines_for_standard_grains) == 0:
            raise SemanticManifestConfigurationError(
                f"This query requires a time spine with granularity {smallest_required_standard_grain.name} or smaller, which is not configured. "
                f"The smallest available time spine granularity is {min(time_spine_sources).name}, which is too large."
                "See documentation for how to configure a new time spine: https://docs.getdbt.com/docs/build/metricflow-time-spine"
            )

        # If the standard grains can't be satisfied by the same time spines as the custom grains, add the largest compatible one.
        if not required_time_spines.intersection(set(compatible_time_spines_for_standard_grains.values())):
            required_time_spines.add(time_spine_sources[max(compatible_time_spines_for_standard_grains)])

        return tuple(sorted(required_time_spines, key=lambda x: x.base_granularity.to_int()))

    @property
    def data_set_description(self) -> str:
        """Description to be displayed when this time spine is used in a data set."""
        return f"Read From Time Spine '{self.sql_table.table_name}'"

    @property
    def custom_grain_names(self) -> Sequence[str]:
        """Names of custom grains defined in this time spine."""
        return tuple(custom_granularity.name for custom_granularity in self.custom_granularities)



from __future__ import annotations

import logging
from collections import defaultdict
from enum import Enum
from functools import cached_property
from typing import Iterable, Mapping, Optional, Sequence

from dbt_semantic_interfaces.protocols import Metric, SemanticManifest, SemanticModel
from dbt_semantic_interfaces.type_enums import MetricType, TimeGranularity
from more_itertools import peekable
from typing_extensions import override

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple, Pair, T
from metricflow_semantics.collection_helpers.syntactic_sugar import mf_first_item, mf_flatten
from metricflow_semantics.experimental.dsi.model_object_lookup import (
    ModelObjectLookup,
)
from metricflow_semantics.experimental.dsi.simple_metric_model_object_lookup import SimpleMetricModelObjectLookup
from metricflow_semantics.experimental.metricflow_exception import MetricFlowInternalError
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, MutableOrderedSet, OrderedSet
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.mf_logging.attribute_pretty_format import AttributeMapping, AttributePrettyFormattable
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantics.simple_metric_input import SimpleMetricInput
from metricflow_semantics.time.granularity import ExpandedTimeGranularity
from metricflow_semantics.time.time_spine_source import TimeSpineSource

logger = logging.getLogger(__name__)


class ManifestObjectLookup(AttributePrettyFormattable):
    """Helps retrieve semantic-manifest objects (i.e. ones from `dbt-semantic-interfaces`).

    These streamlined / minimal lookups were added to use for initializing the semantic graph as the current lookup
    classes have significant initialization times relative to initialization time of the semantic-graph-based resolver.

    Using an example high-complexity semantic manifest, initialization of the current lookups took ~0.2s.
    Initialization of the semantic-graph resolver with the same manifest was < 1s, so using the current lookups would
    have resulted in a significant relative increase.

    Initialization of this lookup and an example call is 100x faster (see `test_model_lookup_performance`).

    Future work includes further refinement of these classes and replacing existing `*Lookup` classes with the ones in
    this module / other implementations.
    """

    def __init__(self, semantic_manifest: SemanticManifest) -> None:  # noqa: D107
        self._semantic_manifest = semantic_manifest
        self._time_spine_sources = TimeSpineSource.build_standard_time_spine_sources(semantic_manifest)
        self._custom_grains = TimeSpineSource.build_custom_granularities(self._time_spine_sources.values())

    @cached_property
    def semantic_manifest(self) -> SemanticManifest:  # noqa: D102
        return self._semantic_manifest

    @cached_property
    def semantic_models(self) -> Sequence[SemanticModel]:  # noqa: D102
        return self._semantic_manifest.semantic_models

    @cached_property
    def _semantic_model_and_simple_metrics_pairs(self) -> Sequence[Pair[SemanticModel, Sequence[Metric]]]:
        model_name_to_simple_metrics: defaultdict[str, list[Metric]] = defaultdict(list)

        for metric in self._semantic_manifest.metrics:
            metric_type = metric.type
            if metric_type is MetricType.SIMPLE:
                metric_aggregation_params = metric.type_params.metric_aggregation_params
                if metric_aggregation_params is None:
                    raise MetricFlowInternalError(
                        LazyFormat("A simple metric is missing `metric_aggregation_params`", metric=metric)
                    )
                model_name_to_simple_metrics[metric_aggregation_params.semantic_model].append(metric)

        semantic_model_and_simple_metrics_pairs: list[Pair[SemanticModel, Sequence[Metric]]] = []
        for semantic_model in self._semantic_manifest.semantic_models:
            semantic_model_and_simple_metrics_pairs.append(
                (semantic_model, model_name_to_simple_metrics.get(semantic_model.name) or ())
            )
        return semantic_model_and_simple_metrics_pairs

    @cached_property
    def simple_metric_model_lookups(self) -> AnyLengthTuple[SimpleMetricModelObjectLookup]:
        """Returns lookups corresponding to semantic models that are associated with simple metrics."""
        return tuple(
            SimpleMetricModelObjectLookup(semantic_model, simple_metrics=simple_metrics)
            for semantic_model, simple_metrics in self._semantic_model_and_simple_metrics_pairs
            if len(simple_metrics) > 0
        )

    @cached_property
    def simple_metric_exclusive_model_lookups(self) -> AnyLengthTuple[ModelObjectLookup]:
        """Returns lookups corresponding to semantic models that are not associated with simple metrics."""
        return tuple(
            ModelObjectLookup(semantic_model)
            for semantic_model, simple_metrics in self._semantic_model_and_simple_metrics_pairs
            if len(simple_metrics) == 0
        )

    @cached_property
    def model_object_lookups(self) -> AnyLengthTuple[ModelObjectLookup]:
        """Return lookups for all semantic models."""
        return self.simple_metric_model_lookups + self.simple_metric_exclusive_model_lookups

    @cached_property
    def model_id_to_lookup(self) -> Mapping[SemanticModelId, ModelObjectLookup]:  # noqa: D102
        return {lookup.model_id: lookup for lookup in self.model_object_lookups}

    @cached_property
    def model_id_to_simple_metric_model_lookup(  # noqa: D102
        self,
    ) -> Mapping[SemanticModelId, SimpleMetricModelObjectLookup]:
        return {lookup.model_id: lookup for lookup in self.simple_metric_model_lookups}

    @cached_property
    def simple_metric_name_to_input(self) -> Mapping[str, SimpleMetricInput]:  # noqa: D102
        return {
            simple_metric_input.name: simple_metric_input
            for lookup in self.simple_metric_model_lookups
            for simple_metric_inputs in lookup.aggregation_configuration_to_simple_metric_inputs.values()
            for simple_metric_input in simple_metric_inputs
        }

    @cached_property
    def entity_name_to_model_lookups(self) -> Mapping[str, OrderedSet[ModelObjectLookup]]:
        """Mapping from the entity name to the model lookups that have the entity."""
        entity_name_to_model_lookups: dict[str, MutableOrderedSet[ModelObjectLookup]] = defaultdict(MutableOrderedSet)
        for model_id, lookup in self.model_id_to_lookup.items():
            for entity in lookup.semantic_model.entities:
                entity_name_to_model_lookups[entity.name].add(lookup)
        return entity_name_to_model_lookups

    @cached_property
    def entity_name_to_model_ids(self) -> Mapping[str, OrderedSet[SemanticModelId]]:
        """Mapping from the entity name to the IDs of the semantic models that contain it."""
        return {
            entity_name: FrozenOrderedSet(model_lookup.model_id for model_lookup in model_lookups)
            for entity_name, model_lookups in self.entity_name_to_model_lookups.items()
        }

    def get_metric(self, metric_name: str) -> Metric:  # noqa: D102
        return self._lookup_object(
            value_type=_ValueType.METRIC,
            name=metric_name,
            name_to_object_mapping=self._metric_name_to_metric,
        )

    def get_metrics(self) -> Iterable[Metric]:  # noqa: D102
        return self._metric_name_to_metric.values()

    @cached_property
    def min_time_grain_in_time_spine(self) -> TimeGranularity:
        """Return the smallest time grain as configured in the time spine."""
        return mf_first_item(sorted(self._time_spine_sources.keys()))

    @cached_property
    def min_time_grain_used_in_models(self) -> Optional[TimeGranularity]:
        """Return the smallest time grain that's used to define a time dimension."""
        time_grains = mf_flatten(
            model_object_lookup.time_dimension_name_to_grain.values()
            for model_object_lookup in self.model_object_lookups
        )
        peekable_grains = peekable(time_grains)
        try:
            peekable_grains.peek()
        except StopIteration:
            return None

        return min(peekable_grains)

    @cached_property
    def expanded_time_grains(self) -> AnyLengthTuple[ExpandedTimeGranularity]:
        """Return the expanded time grains as configured in the time spine."""
        return tuple(self._custom_grains.values())

    def _lookup_object(self, value_type: _ValueType, name: str, name_to_object_mapping: Mapping[str, T]) -> T:
        """Helper method to look up an object in a mapping and raise a helpful error if it is not found."""
        try:
            return name_to_object_mapping[name]
        except KeyError as e:
            raise KeyError(
                LazyFormat(
                    "An object with the given name is not known",
                    value_type=value_type,
                    name=name,
                    known_names=list(name_to_object_mapping.keys()),
                )
            ) from e

    @cached_property
    def _metric_name_to_metric(self) -> Mapping[str, Metric]:
        metric_name_to_metric: dict[str, Metric] = {}

        for metric in self._semantic_manifest.metrics:
            metric_name_to_metric[metric.name] = metric
        return metric_name_to_metric

    @cached_property
    @override
    def _attribute_mapping(self) -> AttributeMapping:
        return dict(
            **super()._attribute_mapping,
            **{
                attribute_name: getattr(self, attribute_name)
                for attribute_name in (
                    "model_object_lookups",
                    "min_time_grain_in_time_spine",
                    "min_time_grain_used_in_models",
                    "expanded_time_grains",
                )
            },
        )


class _ValueType(Enum):
    """Different types of objects stored as values in lookup dictionaries."""

    ENTITY = "entity"
    METRIC = "metric"
    MODEL_ID = "model_id"



from __future__ import annotations

import logging
from functools import cached_property
from typing import Optional

from dbt_semantic_interfaces.references import SemanticModelReference
from typing_extensions import override

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.mf_graph.comparable import Comparable, ComparisonKey
from metricflow_semantics.experimental.singleton import Singleton
from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable
from metricflow_semantics.mf_logging.pretty_formatter import PrettyFormatContext

logger = logging.getLogger(__name__)


@fast_frozen_dataclass(order=False)
class SemanticModelId(Singleton, MetricFlowPrettyFormattable, Comparable):
    """Singleton replacement for `SemanticModelReference`.

    This is used as `SemanticModelReference` is defined in `dbt-semantic-interfaces` and is difficult to change. This
    may be replaced once that's done.
    """

    model_name: str

    @classmethod
    def get_instance(cls, model_name: str) -> SemanticModelId:  # noqa: D102
        return cls._get_instance(model_name=model_name)

    @override
    @cached_property
    def comparison_key(self) -> ComparisonKey:
        return (self.model_name,)

    @override
    def __str__(self) -> str:
        return self.model_name

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        return self.model_name

    @cached_property
    def semantic_model_reference(self) -> SemanticModelReference:
        """In place for migration."""
        return SemanticModelReference(semantic_model_name=self.model_name)



from __future__ import annotations

import logging
from collections.abc import Sequence
from functools import cached_property
from typing import Mapping, Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.references import SemanticModelReference
from typing_extensions import override

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.collection_helpers.syntactic_sugar import mf_flatten
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, MutableOrderedSet
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_trie import DunderNameTrie
from metricflow_semantics.model.semantics.element_filter import GroupByItemSetFilter
from metricflow_semantics.model.semantics.linkable_element_set_base import AnnotatedSpec, BaseGroupByItemSet
from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern

logger = logging.getLogger(__name__)


@fast_frozen_dataclass()
class GroupByItemSet(BaseGroupByItemSet, SerializableDataclass):
    """Implementation of `BaseGroupByItemSet` based on `AnnotatedSpec`."""

    annotated_specs: Tuple[AnnotatedSpec, ...] = ()

    @staticmethod
    def create(  # noqa: D102
        *annotated_specs: AnnotatedSpec,
    ) -> GroupByItemSet:
        dunder_name_to_annotated_spec: dict[str, AnnotatedSpec] = {}
        for annotated_spec in annotated_specs:
            dunder_name = annotated_spec.spec.dunder_name
            existing_annotated_spec = dunder_name_to_annotated_spec.get(dunder_name)
            if existing_annotated_spec is None:
                dunder_name_to_annotated_spec[dunder_name] = annotated_spec
            else:
                dunder_name_to_annotated_spec[dunder_name] = existing_annotated_spec.merge(annotated_spec)

        return GroupByItemSet(
            annotated_specs=tuple(dunder_name_to_annotated_spec.values()),
        )

    @staticmethod
    def create_from_trie(*dunder_name_tries: DunderNameTrie) -> GroupByItemSet:  # noqa: D102
        return GroupByItemSet(
            annotated_specs=tuple(
                annotated_spec
                for annotated_spec in mf_flatten(
                    AnnotatedSpec.create_from_indexed_dunder_name(indexed_dunder_name, descriptor)
                    for trie in dunder_name_tries
                    for indexed_dunder_name, descriptor in trie.name_items()
                )
            )
        )

    @staticmethod
    def create_from_mapping(dunder_name_to_annotated_spec: dict[str, AnnotatedSpec]) -> GroupByItemSet:  # noqa: D102
        return GroupByItemSet(annotated_specs=tuple(dunder_name_to_annotated_spec.values()))

    @cached_property
    def dunder_name_to_annotated_spec(self) -> Mapping[str, AnnotatedSpec]:
        """Return a mapping from the dunder name to the annotated spec."""
        return {annotated_spec.spec.dunder_name: annotated_spec for annotated_spec in self.annotated_specs}

    @override
    def intersection(self, *other_element_sets: GroupByItemSet) -> GroupByItemSet:
        if len(other_element_sets) == 0:
            return self

        common_keys = MutableOrderedSet(self.dunder_name_to_annotated_spec.keys())
        common_keys.intersection_update(
            *(element_set.dunder_name_to_annotated_spec.keys() for element_set in other_element_sets)
        )

        intersected_dunder_name_to_annotated_spec: dict[str, AnnotatedSpec] = {}
        for common_key in common_keys:
            annotated_spec = self.dunder_name_to_annotated_spec[common_key]
            for other_element_set in other_element_sets:
                annotated_spec = annotated_spec.merge(other_element_set.dunder_name_to_annotated_spec[common_key])
            intersected_dunder_name_to_annotated_spec[common_key] = annotated_spec

        return GroupByItemSet.create_from_mapping(intersected_dunder_name_to_annotated_spec)

    @override
    def union(self, *others: GroupByItemSet) -> GroupByItemSet:
        new_dunder_name_to_annotated_spec: dict[str, AnnotatedSpec] = dict(**self.dunder_name_to_annotated_spec)

        for other in others:
            for dunder_name, annotated_spec in other.dunder_name_to_annotated_spec.items():
                current_annotated_spec = new_dunder_name_to_annotated_spec.get(dunder_name)
                if current_annotated_spec is None:
                    new_dunder_name_to_annotated_spec[dunder_name] = annotated_spec
                else:
                    new_dunder_name_to_annotated_spec[dunder_name] = current_annotated_spec.merge(annotated_spec)

                new_dunder_name_to_annotated_spec[dunder_name] = annotated_spec

        return GroupByItemSet.create_from_mapping(new_dunder_name_to_annotated_spec)

    @override
    def filter(self, element_filter: GroupByItemSetFilter) -> GroupByItemSet:
        element_name_allowlist = element_filter.element_name_allowlist
        any_properties_denylist = element_filter.any_properties_denylist

        new_specs: list[AnnotatedSpec] = []
        for annotated_spec in self.annotated_specs:
            if element_name_allowlist is not None and annotated_spec.element_name not in element_name_allowlist:
                continue

            property_set = annotated_spec.property_set

            if not element_filter.any_properties_allowlist.intersection(annotated_spec.property_set):
                continue

            if any_properties_denylist and len(any_properties_denylist.intersection(property_set)) > 0:
                continue

            new_specs.append(annotated_spec)

        return GroupByItemSet(annotated_specs=tuple(new_specs))

    @override
    @property
    def is_empty(self) -> bool:
        return len(self.annotated_specs) == 0

    @override
    @cached_property
    def specs(self) -> AnyLengthTuple[LinkableInstanceSpec]:
        return tuple(annotated_spec.spec for annotated_spec in self.annotated_specs)

    @override
    def filter_by_spec_patterns(self, spec_patterns: Sequence[SpecPattern]) -> GroupByItemSet:
        if len(spec_patterns) == 0:
            return self

        spec_to_annotated_spec: dict[InstanceSpec, AnnotatedSpec] = {
            annotated_spec.spec: annotated_spec for annotated_spec in self.annotated_specs
        }
        specs = tuple(spec_to_annotated_spec.keys())
        for spec_pattern in spec_patterns:
            matched_specs = tuple(spec_pattern.match(specs))
            specs = matched_specs

        return GroupByItemSet(annotated_specs=tuple(spec_to_annotated_spec[matched_spec] for matched_spec in specs))

    @override
    @cached_property
    def derived_from_semantic_models(self) -> Sequence[SemanticModelReference]:
        return tuple(
            FrozenOrderedSet(
                mf_flatten(annotated_spec.derived_from_semantic_models for annotated_spec in self.annotated_specs)
            )
        )



from __future__ import annotations

import typing
from dataclasses import dataclass
from typing import Optional, Sequence, Tuple

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.sql.sql_exprs import SqlExpressionNode
from metricflow_semantics.sql.sql_join_type import SqlJoinType
from metricflow_semantics.visitor import VisitorOutputT
from typing_extensions import override

from metricflow.sql.sql_plan import SqlPlanNode, SqlPlanNodeVisitor, SqlSelectColumn
from metricflow.sql.sql_table_node import SqlTableNode

if typing.TYPE_CHECKING:
    from metricflow.sql.sql_cte_node import SqlCteAliasMapping, SqlCteNode


@dataclass(frozen=True)
class SqlJoinDescription:
    """Describes how sources should be joined together."""

    # The source that goes on the right side of the JOIN keyword.
    right_source: SqlPlanNode
    right_source_alias: str
    join_type: SqlJoinType
    on_condition: Optional[SqlExpressionNode] = None

    def with_right_source(self, new_right_source: SqlPlanNode) -> SqlJoinDescription:
        """Return a copy of this but with the right source replaced."""
        return SqlJoinDescription(
            right_source=new_right_source,
            right_source_alias=self.right_source_alias,
            join_type=self.join_type,
            on_condition=self.on_condition,
        )


@dataclass(frozen=True)
class SqlOrderByDescription:  # noqa: D101
    expr: SqlExpressionNode
    desc: bool


@dataclass(frozen=True, eq=False)
class SqlSelectStatementNode(SqlPlanNode):
    """Represents an SQL Select statement.

    Attributes:
        select_columns: The columns to select.
        from_source: The source of the data for the select statement.
        from_source_alias: Alias for the from source.
        join_descs: Descriptions of the joins to perform.
        group_bys: The columns to group by.
        order_bys: The columns to order by.
        where: The where clause expression.
        limit: The limit of the number of rows to return.
        distinct: Whether the select statement should return distinct rows.
    """

    _description: str
    select_columns: Tuple[SqlSelectColumn, ...]
    from_source: SqlPlanNode
    from_source_alias: str
    cte_sources: Tuple[SqlCteNode, ...]
    join_descs: Tuple[SqlJoinDescription, ...]
    group_bys: Tuple[SqlSelectColumn, ...]
    order_bys: Tuple[SqlOrderByDescription, ...]
    where: Optional[SqlExpressionNode]
    limit: Optional[int]
    distinct: bool

    @staticmethod
    def create(  # noqa: D102
        description: str,
        select_columns: Tuple[SqlSelectColumn, ...],
        from_source: SqlPlanNode,
        from_source_alias: str,
        cte_sources: Tuple[SqlCteNode, ...] = (),
        join_descs: Tuple[SqlJoinDescription, ...] = (),
        group_bys: Tuple[SqlSelectColumn, ...] = (),
        order_bys: Tuple[SqlOrderByDescription, ...] = (),
        where: Optional[SqlExpressionNode] = None,
        limit: Optional[int] = None,
        distinct: bool = False,
    ) -> SqlSelectStatementNode:
        parent_nodes = (from_source,) + tuple(x.right_source for x in join_descs) + cte_sources
        return SqlSelectStatementNode(
            parent_nodes=parent_nodes,
            _description=description,
            select_columns=select_columns,
            from_source=from_source,
            from_source_alias=from_source_alias,
            cte_sources=cte_sources,
            join_descs=join_descs,
            group_bys=group_bys,
            order_bys=order_bys,
            where=where,
            limit=limit,
            distinct=distinct,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_PLAN_SELECT_STATEMENT_ID_PREFIX

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return (
            tuple(super().displayed_properties)
            + tuple(DisplayedProperty(f"col{i}", column) for i, column in enumerate(self.select_columns))
            + (DisplayedProperty("from_source", self.from_source),)
            + tuple(DisplayedProperty(f"join_{i}", join_desc) for i, join_desc in enumerate(self.join_descs))
            + tuple(DisplayedProperty(f"group_by{i}", group_by) for i, group_by in enumerate(self.group_bys))
            + (DisplayedProperty("where", self.where),)
            + tuple(DisplayedProperty(f"order_by{i}", order_by) for i, order_by in enumerate(self.order_bys))
            + (DisplayedProperty("distinct", self.distinct),)
        )

    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_select_statement_node(self)

    @property
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:  # noqa: D102
        return self

    @property
    @override
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        return None

    @property
    @override
    def description(self) -> str:
        return self._description

    @override
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        return self.select_columns

    @override
    def copy(self) -> SqlSelectStatementNode:
        return SqlSelectStatementNode.create(
            description=self._description,
            select_columns=self.select_columns,
            from_source=self.from_source.copy(),
            from_source_alias=self.from_source_alias,
            cte_sources=tuple(node.copy() for node in self.cte_sources),
            join_descs=tuple(
                join_desc.with_right_source(join_desc.right_source.copy()) for join_desc in self.join_descs
            ),
            group_bys=self.group_bys,
            order_bys=self.order_bys,
            where=self.where,
            limit=self.limit,
            distinct=self.distinct,
        )



from __future__ import annotations

import logging
from contextlib import contextmanager
from typing import Iterator, List, Optional, Sequence, Tuple

from typing_extensions import override


class RecordingLogHandler(logging.Handler):
    """A log-record handler that stores them so that they can be checked in tests."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__()
        self._log_records: List[logging.LogRecord] = []

    @override
    def emit(self, record: logging.LogRecord) -> None:
        self._log_records.append(record)

    @property
    def log_records(self) -> Sequence[logging.LogRecord]:
        """Return the log records seen by the handler."""
        return self._log_records

    def get_last_message(self) -> Optional[str]:
        """Return the message in the last log record, or None if this hasn't seen any."""
        if len(self._log_records) == 0:
            return None

        return self._log_records[-1].message


@contextmanager
def recorded_logging_context(
    logger: logging.Logger, logging_level_int: int
) -> Iterator[Tuple[logging.Logger, RecordingLogHandler]]:
    """Context with a logger (with the given log level) and associated handler to check what was logged.

    The handler records all log records emitted that is appropriate for the given level during this context.
    Log propagation could be disabled in this context to clean test log output, but some issues need to be resolved.
    """
    previous_logging_level = logger.level
    handler = RecordingLogHandler()
    logger.addHandler(handler)
    try:
        logger.setLevel(logging_level_int)
        yield logger, handler
    finally:
        logger.removeHandler(handler)
        logger.setLevel(previous_logging_level)



from __future__ import annotations

import logging
from typing import List, Sequence, Set

from dbt_semantic_interfaces.protocols import WhereFilterIntersection
from dbt_semantic_interfaces.references import MetricReference
from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import MetricFlowQueryResolutionIssueSet
from metricflow_semantics.query.issues.parsing.duplicate_column_names import DuplicateOutputColumnIssue
from metricflow_semantics.query.validation_rules.base_validation_rule import PostResolutionQueryValidationRule
from metricflow_semantics.specs.dunder_column_association_resolver import DunderColumnAssociationResolver

logger = logging.getLogger(__name__)


class UniqueOutputColumnValidationRule(PostResolutionQueryValidationRule):
    """Validates that a query does not include the same output column multiple times."""

    @override
    def validate_complex_metric_in_resolution_dag(
        self,
        metric_reference: MetricReference,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        return MetricFlowQueryResolutionIssueSet.empty_instance()

    @override
    def validate_query_in_resolution_dag(
        self,
        metrics_in_query: Sequence[MetricReference],
        where_filter_intersection: WhereFilterIntersection,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        column_association_resolver = DunderColumnAssociationResolver()

        output_column_names: Set[str] = set()
        duplicate_column_names: List[str] = []
        for spec in self._resolve_metric_result.metric_specs + self._resolve_group_by_item_result.group_by_item_specs:
            output_column_name = column_association_resolver.resolve_spec(spec).column_name
            if output_column_name in output_column_names:
                duplicate_column_names.append(output_column_name)
            output_column_names.add(output_column_name)

        if duplicate_column_names:
            return MetricFlowQueryResolutionIssueSet.from_issue(
                DuplicateOutputColumnIssue.from_parameters(
                    duplicate_column_names=duplicate_column_names, query_resolution_path=resolution_path
                ),
            )

        return MetricFlowQueryResolutionIssueSet.empty_instance()

    @override
    def validate_simple_metric_in_resolution_dag(
        self,
        metric_reference: MetricReference,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        return MetricFlowQueryResolutionIssueSet.empty_instance()



from __future__ import annotations

import logging
from collections.abc import Mapping
from functools import cached_property

from dbt_semantic_interfaces.type_enums import DatePart, TimeGranularity
from typing_extensions import override

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe_step import (
    AttributeRecipeStep,
)
from metricflow_semantics.experimental.semantic_graph.builder.subgraph_generator import (
    SemanticSubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.edges.sg_edges import (
    EntityAttributeEdge,
)
from metricflow_semantics.experimental.semantic_graph.nodes.attribute_nodes import (
    TimeAttributeNode,
)
from metricflow_semantics.experimental.semantic_graph.nodes.entity_nodes import (
    TimeNode,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraphEdge,
)
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

logger = logging.getLogger(__name__)


class TimeEntitySubgraphGenerator(SemanticSubgraphGenerator):
    """Generates the time-entity subgraph.

    * The time-entity subgraph consists of the time-entity node with edges to various  attribute
      nodes. (e.g. `day`, `dow`, `custom_year`).
    * An alternative approach would be to use a virtual semantic-model and have time be a configured entity.
    * There is a bug in the current resolver that shows time gain of the defined time dimension as available to
      query even if it is smaller than the smallest grain defined in the time spine.
    * To replicate this behavior, this adds time attribute nodes based on the smallest time grain used in the semantic
      manifest.
    * This behavior is replicated to ensure snapshot consistency with the current resolver but should be addressed
      post-migration.
    """

    @override
    def add_edges_for_manifest(self, edge_list: list[SemanticGraphEdge]) -> None:
        min_time_grain_in_models = self._manifest_object_lookup.min_time_grain_used_in_models
        min_time_grains = [self._manifest_object_lookup.min_time_grain_in_time_spine]
        if min_time_grain_in_models is not None:
            min_time_grains.append(min_time_grain_in_models)
        self._add_edges_for_time_entity_subgraph(
            min_time_grain=min(min_time_grains, key=lambda time_grain: time_grain.to_int()),
            edge_list=edge_list,
        )

    def _add_edges_for_time_entity_subgraph(
        self, min_time_grain: TimeGranularity, edge_list: list[SemanticGraphEdge]
    ) -> None:
        time_entity_node = TimeNode.get_instance()

        # Add an edge from the time-entity node to all queryable grains using the manifest.
        for time_grain in self._time_grain_to_queryable_time_grains[min_time_grain]:
            edge_list.append(
                EntityAttributeEdge.create(
                    tail_node=time_entity_node,
                    head_node=TimeAttributeNode.get_instance_for_time_grain(time_grain),
                    recipe_step=AttributeRecipeStep(
                        set_time_grain_access=ExpandedTimeGranularity(
                            name=time_grain.value, base_granularity=time_grain
                        ),
                    ),
                )
            )
        # Add all applicable date parts.
        for queryable_date_part in self._time_grain_to_applicable_date_parts[min_time_grain]:
            attribute_node = TimeAttributeNode.get_instance_for_date_part(queryable_date_part)
            edge_list.append(
                EntityAttributeEdge.create(
                    tail_node=time_entity_node,
                    head_node=attribute_node,
                    recipe_step=AttributeRecipeStep(
                        add_properties=(GroupByItemProperty.DATE_PART,),
                        set_date_part_access=queryable_date_part,
                    ),
                )
            )

        # Add all applicable custom grains.
        for expanded_time_grain in self._manifest_object_lookup.expanded_time_grains:
            if expanded_time_grain.base_granularity.to_int() >= min_time_grain.to_int():
                attribute_node = TimeAttributeNode.get_instance_for_expanded_time_grain(expanded_time_grain)
                edge_list.append(
                    EntityAttributeEdge.create(
                        tail_node=time_entity_node,
                        head_node=attribute_node,
                        recipe_step=AttributeRecipeStep(
                            add_properties=(GroupByItemProperty.DERIVED_TIME_GRANULARITY,),
                            set_time_grain_access=expanded_time_grain,
                        ),
                    )
                )

    @cached_property
    def _time_grain_to_queryable_time_grains(self) -> Mapping[TimeGranularity, AnyLengthTuple[TimeGranularity]]:
        return {
            min_referenced_time_grain: tuple(
                time_grain
                for time_grain in TimeGranularity
                if time_grain.to_int() >= min_referenced_time_grain.to_int()
            )
            for min_referenced_time_grain in TimeGranularity
        }

    @cached_property
    def _time_grain_to_applicable_date_parts(self) -> Mapping[TimeGranularity, AnyLengthTuple[DatePart]]:
        return {
            time_grain: tuple(date_part for date_part in DatePart if date_part.to_int() >= time_grain.to_int())
            for time_grain in TimeGranularity
        }



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass

from metricflow_semantics.specs.instance_spec import InstanceSpec


@dataclass(frozen=True)
class OrderBySpec(SerializableDataclass):  # noqa: D101
    instance_spec: InstanceSpec
    descending: bool

    def with_alias(self, alias: Optional[str]) -> OrderBySpec:
        """Return a order by spec that's the same as self but with alias replaced."""
        return OrderBySpec(instance_spec=self.instance_spec.with_alias(alias), descending=self.descending)



from __future__ import annotations

import logging
from functools import cached_property

from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest

from metricflow_semantics.experimental.dsi.manifest_object_lookup import ManifestObjectLookup
from metricflow_semantics.experimental.mf_graph.path_finding.pathfinder import MetricFlowPathfinder
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.recipe_writer_path import (
    AttributeRecipeWriterPath,
)
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.sg_linkable_spec_resolver import (
    SemanticGraphGroupByItemSetResolver,
)
from metricflow_semantics.experimental.semantic_graph.builder.graph_builder import SemanticGraphBuilder
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import SemanticGraphEdge, SemanticGraphNode
from metricflow_semantics.model.semantics.metric_lookup import MetricLookup
from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow_semantics.time.time_spine_source import TimeSpineSource

logger = logging.getLogger(__name__)


class SemanticManifestLookup:
    """Provides convenient lookup methods to get semantic attributes for a manifest."""

    def __init__(self, semantic_manifest: SemanticManifest) -> None:
        """Initializer.

        Args:
            semantic_manifest: The semantic manifest for lookups.
        """
        self._semantic_manifest = semantic_manifest
        self._time_spine_sources = TimeSpineSource.build_standard_time_spine_sources(semantic_manifest)
        self.custom_granularities = TimeSpineSource.build_custom_granularities(list(self._time_spine_sources.values()))
        self._semantic_model_lookup = SemanticModelLookup(
            semantic_manifest=semantic_manifest, custom_granularities=self.custom_granularities
        )

        pathfinder = MetricFlowPathfinder[SemanticGraphNode, SemanticGraphEdge, AttributeRecipeWriterPath]()
        self._manifest_object_lookup = ManifestObjectLookup(semantic_manifest)
        graph_builder = SemanticGraphBuilder(manifest_object_lookup=self._manifest_object_lookup)
        semantic_graph = graph_builder.build()

        group_by_item_set_resolver = SemanticGraphGroupByItemSetResolver(
            manifest_object_lookup=self._manifest_object_lookup,
            semantic_graph=semantic_graph,
            path_finder=pathfinder,
        )

        self._metric_lookup = MetricLookup(
            semantic_manifest=semantic_manifest,
            group_by_item_set_resolver=group_by_item_set_resolver,
            manifest_object_lookup=self._manifest_object_lookup,
        )

    @property
    def semantic_manifest(self) -> SemanticManifest:  # noqa: D102
        return self._semantic_manifest

    @property
    def semantic_model_lookup(self) -> SemanticModelLookup:  # noqa: D102
        return self._semantic_model_lookup

    @property
    def metric_lookup(self) -> MetricLookup:  # noqa: D102
        return self._metric_lookup

    @cached_property
    def manifest_object_lookup(self) -> ManifestObjectLookup:  # noqa: D102
        return self._manifest_object_lookup



from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Mapping
from typing import Set

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe import (
    AttributeRecipe,
    IndexedDunderName,
)
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.recipe_writer_weight import (
    AttributeRecipeWriterWeightFunction,
)
from metricflow_semantics.experimental.semantic_graph.nodes.node_labels import (
    ConfiguredEntityLabel,
    JoinedModelLabel,
    KeyAttributeLabel,
    LocalModelLabel,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import SemanticGraph, SemanticGraphNode
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_descriptor import DunderNameDescriptor
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_trie import (
    DunderNameTrie,
    MutableDunderNameTrie,
)
from metricflow_semantics.helpers.performance_helpers import ExecutionTimer
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantics.linkable_element import LinkableElementType
from metricflow_semantics.model.semantics.semantic_model_join_evaluator import MAX_JOIN_HOPS

logger = logging.getLogger(__name__)


DunderNameTuple = AnyLengthTuple[str]


class EntityKeyTrieResolver:
    """A resolver to figure out the possible group-by items that are entity-keys.

     i.e. the available items when the user wants to `query for an entity`.

    For example, the `bookings` metric could be queried by the following entity keys (if you include entity links):

        booking
        host
        guest
        booking__listing
        listing__user
        ...

    The entity-key trie represents all possible group-by items that represent an entity key.

    The entity-key trie is needed to resolve the available group-by metrics for a query. Group-by metrics
    are modeled as a join using a subquery where the group-by item is the entity key. e.g. for the query (`bookings` by
    `listing__views` ), the query (`bookings` by `listing`) is joined with the subquery (`views` by `listing`) on
    the entity key.
    """

    def __init__(self, semantic_graph: SemanticGraph) -> None:  # noqa: D107
        self._verbose_debug_logs = False
        self._semantic_graph = semantic_graph

    def resolve_entity_key_trie_mapping(self) -> Mapping[SemanticGraphNode, DunderNameTrie]:
        """Returns a mapping from the local-model to the associated entity-key trie.

        This is returned as single mapping because to resolve the group-by metrics for any measure, you need them all
        anyway.
        """
        with ExecutionTimer() as resolution_timer:
            result = self._resolve_entity_key_trie_mapping()
        logger.info(LazyFormat("Resolved entity-key trie mapping", duration=resolution_timer.total_duration))
        return result

    def _resolve_entity_key_trie_mapping(self) -> Mapping[SemanticGraphNode, DunderNameTrie]:
        local_model_nodes = self._semantic_graph.nodes_with_labels(LocalModelLabel.get_instance())
        key_attribute_nodes = self._semantic_graph.nodes_with_labels(KeyAttributeLabel.get_instance())
        joined_model_nodes = self._semantic_graph.nodes_with_labels(JoinedModelLabel.get_instance())
        configured_entity_nodes = self._semantic_graph.nodes_with_labels(ConfiguredEntityLabel.get_instance())

        allowed_nodes: Set[SemanticGraphNode] = set()
        allowed_nodes.update(local_model_nodes, key_attribute_nodes, joined_model_nodes, configured_entity_nodes)

        source_nodes = local_model_nodes
        target_nodes = key_attribute_nodes
        max_entity_links = MAX_JOIN_HOPS

        if self._verbose_debug_logs:
            logger.debug(
                LazyFormat(
                    "Starting path-finding for key queries",
                    source_nodes=source_nodes,
                    target_nodes=target_nodes,
                    allowed_nodes=allowed_nodes,
                )
            )

        # We're looking to find all possible paths from the local-model nodes (the source nodes) to the entity-key
        # nodes (the target nodes).
        # To do this efficiently, this uses a method similar to Dijkstra's algorithm. For each node, we maintain
        # a lookup that records the next possible node on paths that lead to entity-key nodes. By starting from the
        # target nodes (instead of the source nodes), we can ensure that by the end of the exploration, the lookup
        # only contains valid paths from source nodes to target nodes.
        semantic_graph = self._semantic_graph
        notifications_to_process: dict[SemanticGraphNode, list[_FoundPathToNodeNotification]] = defaultdict(list)
        initial_recipe = AttributeRecipe()

        # Since we're starting at the target node and going backwards, initially populate the lookup for the target
        # nodes.
        for node in target_nodes:
            recipe_at_node = initial_recipe.push_step(node.recipe_step_to_append)
            for edge_from_predecessor in semantic_graph.edges_with_head_node(node):
                predecessor_node = edge_from_predecessor.tail_node
                recipe_at_predecessor_node = recipe_at_node.push_steps(
                    edge_from_predecessor.recipe_step_to_append,
                    edge_from_predecessor.tail_node.recipe_step_to_append,
                )
                notification = _FoundPathToNodeNotification(next_node=node, recipe=recipe_at_predecessor_node)

                if self._verbose_debug_logs:
                    logger.debug(
                        LazyFormat(
                            "Adding initial notification",
                            predecessor_node=predecessor_node,
                            notification=notification,
                        )
                    )
                notifications_to_process[predecessor_node].append(notification)

        found_path_to_node_notifications: dict[SemanticGraphNode, list[_FoundPathToNodeNotification]] = defaultdict(
            list
        )

        # At each step, explore the predecessors of the nodes that were discovered in the previous step. Repeat until.
        # there are no more new nodes / new paths discovered.
        logger.info(
            LazyFormat(
                "Starting exploration for entity-key paths",
                source_node_count=len(source_nodes),
                target_node_count=len(target_nodes),
            )
        )
        step_index = 0
        processed_notification_count = 0
        while len(notifications_to_process) > 0:
            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Processing step", step_index=step_index, notifications_to_process=notifications_to_process
                    )
                )

            next_notifications_to_process: dict[SemanticGraphNode, list[_FoundPathToNodeNotification]] = defaultdict(
                list
            )

            for current_node, found_new_path_notifications in notifications_to_process.items():
                processed_notification_count += len(found_new_path_notifications)
                if current_node in source_nodes:
                    found_path_to_node_notifications[current_node].extend(found_new_path_notifications)
                    continue

                for new_path_notification in found_new_path_notifications:
                    recipe_at_current_node = new_path_notification.recipe

                    for edge_from_predecessor in semantic_graph.edges_with_head_node(current_node):
                        predecessor_node = edge_from_predecessor.tail_node

                        tail_node_step = edge_from_predecessor.tail_node.recipe_step_to_append
                        edge_step = edge_from_predecessor.recipe_step_to_append

                        if AttributeRecipeWriterWeightFunction.repeated_dunder_name_elements(
                            recipe=recipe_at_current_node, step=tail_node_step, other_step=edge_step
                        ) or AttributeRecipeWriterWeightFunction.repeated_model_join(
                            recipe=recipe_at_current_node, step=tail_node_step, other_step=edge_step
                        ):
                            continue

                        next_entity_link_count = (
                            len(recipe_at_current_node.entity_link_names)
                            + (1 if tail_node_step.add_entity_link else 0)
                            + (1 if edge_step.add_entity_link else 0)
                        )

                        if next_entity_link_count > max_entity_links:
                            continue

                        # Handle a special case where entities can be queried by an entity link prefix in the same
                        # semantic model, but not when a part of a join.
                        next_model_count = (
                            len(recipe_at_current_node.joined_model_ids)
                            + (1 if tail_node_step.add_model_join else 0)
                            + (1 if edge_step.add_model_join else 0)
                        )
                        if next_model_count > 1 and next_entity_link_count > (next_model_count - 1):
                            continue

                        recipe_at_predecessor = recipe_at_current_node.push_steps(
                            edge_from_predecessor.recipe_step_to_append,
                            edge_from_predecessor.tail_node.recipe_step_to_append,
                        )

                        notification = _FoundPathToNodeNotification(
                            next_node=current_node, recipe=recipe_at_predecessor
                        )
                        if self._verbose_debug_logs:
                            logger.debug(
                                LazyFormat(
                                    "Adding notification",
                                    predecessor_node=predecessor_node,
                                    notification=notification,
                                )
                            )
                        next_notifications_to_process[predecessor_node].append(notification)
            notifications_to_process = next_notifications_to_process
            step_index += 1
        logger.info(
            LazyFormat(
                "Finished exploration for entity-key paths",
                step_index=step_index,
                processed_notification_count=processed_notification_count,
            )
        )

        # Once the above loop has finished, the lookup can be used to figure out the possible path (encoded in the
        # recipe).
        source_node_to_trie: dict[SemanticGraphNode, MutableDunderNameTrie] = defaultdict(MutableDunderNameTrie)
        added_name_item_count = 0
        for source_node, notifications in found_path_to_node_notifications.items():
            items_to_add: list[tuple[IndexedDunderName, DunderNameDescriptor]] = []
            for notification in notifications:
                recipe = notification.recipe
                indexed_dunder_name = recipe.indexed_dunder_name
                last_model_id = recipe.last_model_id
                items_to_add.append(
                    (
                        indexed_dunder_name,
                        DunderNameDescriptor(
                            element_type=LinkableElementType.ENTITY,
                            time_grain=None,
                            date_part=None,
                            element_properties=(),
                            origin_model_ids=(last_model_id,) if last_model_id else (),
                            derived_from_model_ids=notification.recipe.joined_model_ids,
                            entity_key_queries_for_group_by_metric=(),
                        ),
                    )
                )

            source_node_to_trie[source_node].add_name_items(items_to_add)
            added_name_item_count += len(items_to_add)
        logger.info(LazyFormat("Finished mapping source node to entity-key trie", source_node_count=len(source_nodes)))
        return {source_node: source_node_to_trie[source_node] for source_node in source_nodes}


@fast_frozen_dataclass()
class _FoundPathToNodeNotification:
    """The value for the keys in the dict that maps a node to the next node in the path (to a target node)."""

    next_node: SemanticGraphNode
    recipe: AttributeRecipe



from __future__ import annotations

import difflib
import logging
from collections import defaultdict
from collections.abc import Mapping, Sequence

from metricflow_semantics.helpers.table_helpers import IsolatedTabulateRunner
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


class PaddedTextTableBuilder:
    """A builder for a pair of text tables to include congruent padding for comparison in tests.

    The issue with comparing text tables is that if there's mismatch due to a long item in only one of the compared
    tables, the width of the entire corresponding table can change. This results in a diff for the entire table.

    This generates tables that pad columns using appropriate whitespace so that in the above case, the diff only
    includes relevant rows.
    """

    def __init__(self, tablefmt: str = "simple_outline") -> None:  # noqa: D107
        self._column_name_to_max_width: dict[str, int] = defaultdict(int)
        self._left_rows: list[Mapping[str, str]] = []
        self._right_rows: list[Mapping[str, str]] = []
        self._tablefmt = tablefmt

    def _record_max_column_widths(self, rows: Sequence[Mapping[str, str]]) -> None:
        """For each row, update the maximum known width of the column name / value."""
        for row in rows:
            for column_name, value in row.items():
                self._column_name_to_max_width[column_name] = max(
                    self._column_name_to_max_width[column_name], len(value), len(column_name)
                )

    def add_left_rows(self, rows: Sequence[Mapping[str, str]]) -> None:  # noqa: D102
        self._record_max_column_widths(rows)
        self._left_rows.extend(rows)

    def add_right_rows(self, rows: Sequence[Mapping[str, str]]) -> None:  # noqa: D102
        self._record_max_column_widths(rows)
        self._right_rows.extend(rows)

    def format_left_table(self) -> str:  # noqa: D102
        return self._format_table(self._left_rows)

    def format_right_table(self) -> str:  # noqa: D102
        return self._format_table(self._right_rows)

    def _format_table(self, rows: Sequence[Mapping[str, str]]) -> str:
        new_rows: list[dict[str, str]] = []
        for row in rows:
            new_row = {}
            for column_name, max_width in self._column_name_to_max_width.items():
                padded_column_value = (row.get(column_name) or "").ljust(max_width)
                padded_column_name = column_name.ljust(max_width)
                new_row[padded_column_name] = padded_column_value
            new_rows.append(new_row)

        return IsolatedTabulateRunner.tabulate(tabular_data=new_rows, headers="keys", tablefmt=self._tablefmt)

    def assert_tables_equal(self, log_table_on_match: bool) -> None:
        """Assert that the left and right tables match.

        If `log_table_on_match` is set, the generated table is logged (for prototyping).
        """
        left_table = self.format_left_table()
        right_table = self.format_right_table()

        if left_table == right_table:
            additional_format_kwargs = {"table": left_table} if log_table_on_match else {}
            logger.debug(LazyFormat("Left and right tables match", **additional_format_kwargs))
            return

        diff_lines = difflib.unified_diff(
            a=left_table.splitlines(keepends=True),
            b=right_table.splitlines(keepends=True),
            fromfile="Left Table",
            tofile="Right Table",
            n=0,
        )

        diff = "".join(diff_lines)
        assert False, LazyFormat(
            "Mismatch between left and right tables",
            left_table=left_table,
            right_table=right_table,
            diff=diff,
        ).evaluated_value



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import Mapping, Optional, Sequence, Tuple

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.visitor import VisitorOutputT
from typing_extensions import override

from metricflow.sql.sql_plan import SqlPlanNode, SqlPlanNodeVisitor, SqlSelectColumn
from metricflow.sql.sql_select_node import SqlSelectStatementNode
from metricflow.sql.sql_table_node import SqlTableNode


@dataclass(frozen=True, eq=False)
class SqlCteNode(SqlPlanNode):
    """Represents a single common table expression."""

    select_statement: SqlPlanNode
    cte_alias: str

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(select_statement: SqlPlanNode, cte_alias: str) -> SqlCteNode:  # noqa: D102
        return SqlCteNode(
            parent_nodes=(select_statement,),
            select_statement=select_statement,
            cte_alias=cte_alias,
        )

    def with_new_select(self, new_select_statement: SqlPlanNode) -> SqlCteNode:
        """Return a node with the same attributes but with the new SELECT statement."""
        return SqlCteNode.create(
            select_statement=new_select_statement,
            cte_alias=self.cte_alias,
        )

    @override
    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        return visitor.visit_cte_node(self)

    @property
    @override
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:
        return None

    @property
    @override
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        return None

    @property
    @override
    def description(self) -> str:
        return "CTE"

    @classmethod
    @override
    def id_prefix(cls) -> IdPrefix:
        return StaticIdPrefix.SQL_PLAN_COMMON_TABLE_EXPRESSION_ID_PREFIX

    @override
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        return self.select_statement.nearest_select_columns(cte_source_mapping)

    @override
    def copy(self) -> SqlCteNode:
        return SqlCteNode.create(
            select_statement=self.select_statement.copy(),
            cte_alias=self.cte_alias,
        )


@dataclass(frozen=True)
class SqlCteAliasMapping(Mergeable):
    """Thin, dict-like object that maps an alias to the associated `SqlCteNode`.

    When merged, the entries from the right mapping take precedence over the entries from the left.
    """

    cte_alias_to_cte_node_items: Tuple[Tuple[str, SqlCteNode], ...] = ()

    @staticmethod
    def create(cte_alias_to_cte_node_mapping: Mapping[str, SqlCteNode]) -> SqlCteAliasMapping:  # noqa: D102
        cte_alias_to_cte_node_pairs = []
        for cte_alias, cte_node in cte_alias_to_cte_node_mapping.items():
            cte_alias_to_cte_node_pairs.append((cte_alias, cte_node))

        return SqlCteAliasMapping(cte_alias_to_cte_node_items=tuple(cte_alias_to_cte_node_pairs))

    @cached_property
    def _cte_alias_to_cte_node_dict(self) -> Mapping[str, SqlCteNode]:
        return {item[0]: item[1] for item in self.cte_alias_to_cte_node_items}

    def get_cte_node_for_alias(self, cte_alias: str) -> Optional[SqlCteNode]:
        """Return the associated `SqlCteNode` for the given alias, or None if the given alias is not known."""
        return self._cte_alias_to_cte_node_dict.get(cte_alias)

    @override
    def merge(self, other: SqlCteAliasMapping) -> SqlCteAliasMapping:
        new_mapping = dict(self._cte_alias_to_cte_node_dict)
        for cte_alias, cte_node in other.cte_alias_to_cte_node_items:
            new_mapping[cte_alias] = cte_node
        return SqlCteAliasMapping.create(new_mapping)

    @classmethod
    @override
    def empty_instance(cls) -> SqlCteAliasMapping:
        return SqlCteAliasMapping()



from __future__ import annotations

import logging
from collections import defaultdict
from typing import Iterable, Optional

from dbt_semantic_interfaces.references import ElementReference, MetricReference
from typing_extensions import override

from metricflow_semantics.collection_helpers.syntactic_sugar import mf_first_item
from metricflow_semantics.experimental.cache.mf_cache import ResultCache
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.dsi.manifest_object_lookup import ManifestObjectLookup
from metricflow_semantics.experimental.metricflow_exception import MetricFlowInternalError
from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.mf_graph.path_finding.pathfinder import MetricFlowPathfinder
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, MutableOrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.annotated_spec_linkable_element_set import (
    GroupByItemSet,
)
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe import IndexedDunderName
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.recipe_writer_path import (
    AttributeRecipeWriterPath,
)
from metricflow_semantics.experimental.semantic_graph.nodes.node_labels import (
    LocalModelLabel,
    MetricLabel,
    MetricTimeLabel,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraph,
    SemanticGraphEdge,
    SemanticGraphNode,
)
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_descriptor import DunderNameDescriptor
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_trie import (
    DunderNameTrie,
    MutableDunderNameTrie,
)
from metricflow_semantics.experimental.semantic_graph.trie_resolver.group_by_metric_resolver import (
    GroupByMetricTrieResolver,
)
from metricflow_semantics.experimental.semantic_graph.trie_resolver.simple_resolver import SimpleTrieResolver
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantics.element_filter import GroupByItemSetFilter
from metricflow_semantics.model.semantics.linkable_element_set_base import BaseGroupByItemSet
from metricflow_semantics.model.semantics.linkable_spec_resolver import (
    GroupByItemSetResolver,
)

logger = logging.getLogger(__name__)


class SemanticGraphGroupByItemSetResolver(GroupByItemSetResolver):
    """An implementation of `GroupByItemSetResolver` using the semantic graph."""

    def __init__(  # noqa: D107
        self,
        semantic_graph: SemanticGraph,
        manifest_object_lookup: ManifestObjectLookup,
        path_finder: MetricFlowPathfinder[SemanticGraphNode, SemanticGraphEdge, AttributeRecipeWriterPath],
    ) -> None:
        self._semantic_graph = semantic_graph
        self._pathfinder = path_finder
        self._manifest_object_lookup = manifest_object_lookup
        self._simple_resolver = SimpleTrieResolver(semantic_graph, path_finder)
        self._simple_resolver_limit_one_model = SimpleTrieResolver(semantic_graph, path_finder, max_path_model_count=1)
        self._group_by_metric_resolver = GroupByMetricTrieResolver(semantic_graph, path_finder)
        self._local_model_nodes = self._semantic_graph.nodes_with_labels(LocalModelLabel.get_instance())
        self._metric_time_node_set = FrozenOrderedSet(
            (self._semantic_graph.node_with_label(MetricTimeLabel.get_instance()),)
        )

        self._result_cache_for_distinct_values: ResultCache[
            tuple[Optional[GroupByItemSetFilter]], BaseGroupByItemSet
        ] = ResultCache()

        self._result_cache_for_common_set: ResultCache[_CommonSetCacheKey, BaseGroupByItemSet] = ResultCache()

    @override
    def get_set_for_distinct_values_query(
        self, set_filter: Optional[GroupByItemSetFilter] = None
    ) -> BaseGroupByItemSet:
        cache_key = (set_filter,)
        cache_result = self._result_cache_for_distinct_values.get(cache_key)
        if cache_result:
            return cache_result.value
        tries_to_union: list[DunderNameTrie] = []
        for local_model_node in self._local_model_nodes:
            source_nodes = FrozenOrderedSet((local_model_node,))
            local_model_trie = self._simple_resolver_limit_one_model.resolve_trie(
                source_nodes, set_filter
            ).dunder_name_trie
            group_by_metric_trie = self._group_by_metric_resolver.resolve_trie(
                source_nodes, set_filter
            ).dunder_name_trie
            tries_to_union.append(local_model_trie)
            tries_to_union.append(group_by_metric_trie)

        metric_time_trie = self._simple_resolver.resolve_trie(self._metric_time_node_set, set_filter).dunder_name_trie

        # Since `TimeEntitySubgraphGenerator` could add time grain nodes that are finer than the grain of time spine,
        # filter those out.
        int_value_of_min_time_spine_grain = self._manifest_object_lookup.min_time_grain_in_time_spine.to_int()
        filtered_items_from_metric_time_trie: list[tuple[IndexedDunderName, DunderNameDescriptor]] = []

        for indexed_dunder_name, descriptor in metric_time_trie.name_items():
            # Allow `metric_time` at grains >= min grain.
            if (
                descriptor.time_grain is not None
                and descriptor.time_grain.base_granularity.to_int() >= int_value_of_min_time_spine_grain
            ):
                filtered_items_from_metric_time_trie.append((indexed_dunder_name, descriptor))
            # Allow `metric_time` with a date part compatible with the min grain.
            elif (
                descriptor.date_part is not None and descriptor.date_part.to_int() >= int_value_of_min_time_spine_grain
            ):
                filtered_items_from_metric_time_trie.append((indexed_dunder_name, descriptor))

        filtered_metric_time_trie = MutableDunderNameTrie()
        filtered_metric_time_trie.add_name_items(filtered_items_from_metric_time_trie)

        tries_to_union.append(filtered_metric_time_trie)
        result_trie = MutableDunderNameTrie.union_merge_common(tries_to_union)

        return self._result_cache_for_distinct_values.set_and_get(
            cache_key, GroupByItemSet.create_from_trie(result_trie)
        )

    @override
    def get_common_set(
        self,
        metric_references: Iterable[MetricReference] = (),
        set_filter: Optional[GroupByItemSetFilter] = None,
        joins_disallowed: bool = False,
    ) -> BaseGroupByItemSet:
        """Get the group-by items common to all metrics.

        If `joins_disallowed` is set, then only group-by items that can be resolved using the semantic models where the
        simple metrics are defined will be returned.
        """
        label_to_references: defaultdict[MetricFlowGraphLabel, set[ElementReference]] = defaultdict(set)
        for metric_reference in metric_references:
            label_to_references[MetricLabel.get_instance(metric_reference.element_name)].add(metric_reference)

        # Sanity check.
        for label, references in label_to_references.items():
            assert len(references) == 1, LazyFormat(
                "Different references map to the same label.",
                references=references,
                label=label,
            )

        if len(label_to_references) == 0:
            return GroupByItemSet()

        node_labels: FrozenOrderedSet[MetricFlowGraphLabel] = FrozenOrderedSet(sorted(label_to_references))
        cache_key = _CommonSetCacheKey(
            node_labels=node_labels, set_filter=set_filter, joins_disallowed=joins_disallowed
        )
        cached_result = self._result_cache_for_common_set.get(cache_key)
        if cached_result:
            return cached_result.value

        source_nodes: MutableOrderedSet[SemanticGraphNode] = MutableOrderedSet()

        for label in node_labels:
            matching_nodes = self._semantic_graph.nodes_with_labels(label)
            if len(matching_nodes) != 1:
                raise MetricFlowInternalError(
                    LazyFormat(
                        "Did not find exactly 1 node in the semantic graph for the given label",
                        label=label,
                        associated_references=label_to_references.get(label),
                        matching_nodes=matching_nodes,
                        metric_nodes=lambda: self._semantic_graph.nodes_with_labels(MetricLabel.get_instance()),
                    )
                )

            source_nodes.add(mf_first_item(matching_nodes))

        if joins_disallowed:
            simple_trie = self._simple_resolver_limit_one_model.resolve_trie(source_nodes, set_filter).dunder_name_trie
            return self._result_cache_for_common_set.set_and_get(
                cache_key, GroupByItemSet.create_from_trie(simple_trie)
            )

        simple_trie = self._simple_resolver.resolve_trie(source_nodes, set_filter).dunder_name_trie
        group_by_metric_trie = self._group_by_metric_resolver.resolve_trie(source_nodes, set_filter).dunder_name_trie

        return self._result_cache_for_common_set.set_and_get(
            cache_key, GroupByItemSet.create_from_trie(simple_trie, group_by_metric_trie)
        )


@fast_frozen_dataclass()
class _CommonSetCacheKey:
    node_labels: FrozenOrderedSet[MetricFlowGraphLabel]
    set_filter: Optional[GroupByItemSetFilter]
    joins_disallowed: bool



from __future__ import annotations

import dataclasses
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Mapping
from dataclasses import dataclass
from functools import cached_property
from typing import Iterable, Optional, Sequence

from dbt_semantic_interfaces.naming.keywords import DUNDER
from typing_extensions import Self, override

from metricflow_semantics.collection_helpers.dictionaries import mf_common_keys
from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.metricflow_exception import MetricFlowInternalError
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe import IndexedDunderName
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_descriptor import DunderNameDescriptor
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


class DunderNameTrie(ABC):
    """A read-only interface for a trie-based data structure that stores dunder names and associated properties.

    Unlike a typical trie that is based on individual characters, this is based on the names that are joined with `__`.
    For example, the trie that contains

        listing__user
        listing__country

    would be represented as:

        listing -> user
                -> country

    The trie helps to simplify union / intersection operations that are done to generate the list of available group-by
    items for a query.

    There's currently a mutable implementation, but a frozen one may be added if it turns out to be useful.
    """

    @abstractmethod
    def name_items(self, max_length: Optional[int] = None) -> Sequence[tuple[IndexedDunderName, DunderNameDescriptor]]:
        """Return a sequence of tuples that represent all dunder names that are represented by this trie.

        The first item is the indexed dunder-name (e.g. ("listing", "country")) and the second is the descriptor
        associated with that name.
        """
        raise NotImplementedError

    def dunder_names(self) -> Sequence[str]:
        """Return the dunder names that are represented by this trie (e.g. ["listing__country", ...]).

        This is mainly used for tests / debugging.
        """
        return tuple(DUNDER.join(indexed_dunder_name) for indexed_dunder_name, _ in self.name_items())

    @classmethod
    @abstractmethod
    def union_merge_common(cls, trie_sequence: Sequence[Self]) -> Self:
        """Union names from a sequence of tries, and merge the descriptors for common names."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def union_exclude_common(cls, trie_sequence: Sequence[Self]) -> Self:
        """Union names from a sequence of tries, but remove the common ones.

        In this operation, non-unique names are discarded. This is to represent ambiguous join paths which are not
        allowed in the query interface.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def intersection_merge_common(cls, trie_sequence: Sequence[Self]) -> Self:
        """Intersect names from a sequence of tries, and merge the descriptors for common names.

        In an intersection operation, names that do not exist in all tries are discarded. This represents the
        requirement that, for a derived metric, a group-by name must be available for each of the input metrics
        (similar situation for a query of multiple metrics).
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def name_element_to_descriptor(self) -> Mapping[str, DunderNameDescriptor]:
        """Return a mapping of the name elements in this trie to the descriptor.

        For example, this would return `{"listing": Descriptor(...)}` for a trie that stored `listing` and
        `listing__country`. See `name_element_to_next_trie` for retrieval of the `country` element.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def next_name_element_to_trie(self) -> Mapping[str, MutableDunderNameTrie]:
        """Return a mapping from the next name element to the associated trie.

        For example, a trie that represents `listing` and `listing__country` would return
        `{"listing": DunderNameTrie(...)}`.
        """
        raise NotImplementedError

    def mutable_copy(self) -> MutableDunderNameTrie:
        """Return a mutable copy of this trie. Child tries are copied as well.

        Descriptors don't need to be copied as they are immutable.
        """
        return MutableDunderNameTrie(
            name_element_to_descriptor_dict=dict(self.name_element_to_descriptor),
            name_element_to_next_trie_dict={
                name_element: next_trie.mutable_copy()
                for name_element, next_trie in self.next_name_element_to_trie.items()
            },
        )


@dataclass
class MutableDunderNameTrie(DunderNameTrie):
    """A mutable implementation of `DunderNameTrie`."""

    name_element_to_descriptor_dict: dict[str, DunderNameDescriptor] = dataclasses.field(default_factory=dict)
    name_element_to_next_trie_dict: dict[str, MutableDunderNameTrie] = dataclasses.field(
        default_factory=lambda: defaultdict(MutableDunderNameTrie)
    )

    @classmethod
    @override
    def union_merge_common(cls, trie_sequence: Sequence[DunderNameTrie]) -> MutableDunderNameTrie:
        if len(trie_sequence) == 0:
            return MutableDunderNameTrie()
        elif len(trie_sequence) == 1:
            return trie_sequence[0].mutable_copy()

        new_name_element_to_descriptors: dict[str, DunderNameDescriptor] = {}
        for trie in trie_sequence:
            for name_element, descriptor in trie.name_element_to_descriptor.items():
                previous_descriptor = new_name_element_to_descriptors.get(name_element)
                if previous_descriptor is None:
                    new_name_element_to_descriptors[name_element] = descriptor
                else:
                    new_name_element_to_descriptors[name_element] = previous_descriptor.merge(descriptor)

        next_name_element_to_tries_for_union: dict[str, list[MutableDunderNameTrie]] = defaultdict(list)
        for trie in trie_sequence:
            for name_element, next_trie in trie.next_name_element_to_trie.items():
                next_name_element_to_tries_for_union[name_element].append(next_trie)

        new_name_element_to_next_trie: dict[str, MutableDunderNameTrie] = {}
        for name_element, next_tries in next_name_element_to_tries_for_union.items():
            new_name_element_to_next_trie[name_element] = MutableDunderNameTrie.union_merge_common(next_tries)

        return MutableDunderNameTrie(
            name_element_to_descriptor_dict=new_name_element_to_descriptors,
            name_element_to_next_trie_dict=new_name_element_to_next_trie,
        )

    @classmethod
    @override
    def union_exclude_common(cls, trie_sequence: Sequence[DunderNameTrie]) -> MutableDunderNameTrie:
        if len(trie_sequence) == 0:
            return MutableDunderNameTrie()
        elif len(trie_sequence) == 1:
            return trie_sequence[0].mutable_copy()

        common_name_elements_for_descriptors = mf_common_keys(
            tuple(trie.name_element_to_descriptor for trie in trie_sequence)
        )

        new_name_element_to_descriptors: dict[str, DunderNameDescriptor] = {
            name_element: descriptor
            for trie in trie_sequence
            for name_element, descriptor in trie.name_element_to_descriptor.items()
            # if name_element not in new_ambiguous_name_elements
            if name_element not in common_name_elements_for_descriptors
        }

        next_name_element_to_tries_for_union: dict[str, list[MutableDunderNameTrie]] = defaultdict(list)
        for trie in trie_sequence:
            for name_element, next_trie in trie.next_name_element_to_trie.items():
                next_name_element_to_tries_for_union[name_element].append(next_trie)

        new_name_element_to_next_trie: dict[str, MutableDunderNameTrie] = {}
        for name_element, next_tries in next_name_element_to_tries_for_union.items():
            new_name_element_to_next_trie[name_element] = MutableDunderNameTrie.union_exclude_common(next_tries)

        return MutableDunderNameTrie(
            name_element_to_descriptor_dict=new_name_element_to_descriptors,
            name_element_to_next_trie_dict=new_name_element_to_next_trie,
        )

    @classmethod
    @override
    def intersection_merge_common(cls, trie_sequence: Sequence[DunderNameTrie]) -> MutableDunderNameTrie:
        if len(trie_sequence) == 0:
            return MutableDunderNameTrie()
        elif len(trie_sequence) == 1:
            return trie_sequence[0].mutable_copy()

        left_trie = trie_sequence[0]
        other_tries = trie_sequence[1:]

        intersected_name_elements_for_descriptors = mf_common_keys(
            tuple(trie.name_element_to_descriptor for trie in trie_sequence)
        )

        new_element_name_to_descriptor: dict[str, DunderNameDescriptor] = {
            name_element: left_trie.name_element_to_descriptor[name_element]
            for name_element in intersected_name_elements_for_descriptors
        }

        for trie in other_tries:
            for name_element in intersected_name_elements_for_descriptors:
                new_element_name_to_descriptor[name_element] = new_element_name_to_descriptor[name_element].merge(
                    trie.name_element_to_descriptor[name_element]
                )

        intersected_name_elements_for_next_tries = mf_common_keys(
            tuple(trie.next_name_element_to_trie for trie in trie_sequence)
        )

        name_element_to_next_tries: dict[str, list[MutableDunderNameTrie]] = defaultdict(list)
        for trie in trie_sequence:
            for name_element, next_trie in trie.next_name_element_to_trie.items():
                if name_element not in intersected_name_elements_for_next_tries:
                    continue
                name_element_to_next_tries[name_element].append(next_trie)

        new_name_element_to_next_trie = {
            name_element: MutableDunderNameTrie.intersection_merge_common(next_tries)
            for name_element, next_tries in name_element_to_next_tries.items()
        }

        return MutableDunderNameTrie(
            name_element_to_descriptor_dict=new_element_name_to_descriptor,
            name_element_to_next_trie_dict=new_name_element_to_next_trie,
        )

    @cached_property
    @override
    def name_element_to_descriptor(self) -> Mapping[str, DunderNameDescriptor]:
        return self.name_element_to_descriptor_dict

    @cached_property
    @override
    def next_name_element_to_trie(self) -> Mapping[str, MutableDunderNameTrie]:
        return self.name_element_to_next_trie_dict

    def update_derived_from_model_ids(self, derived_from_model_ids: AnyLengthTuple[SemanticModelId]) -> None:
        """Update the descriptor for all represented names to include additional `derived_from_model_ids`."""
        self.name_element_to_descriptor_dict.update(
            {
                name_element: descriptor.merge_derived_from_model_ids(derived_from_model_ids)
                for name_element, descriptor in self.name_element_to_descriptor_dict.items()
            }
        )
        for _, next_trie in self.name_element_to_next_trie_dict.items():
            next_trie.update_derived_from_model_ids(derived_from_model_ids)

    def add_name_items(self, items: Iterable[tuple[IndexedDunderName, DunderNameDescriptor]]) -> None:
        """Build out this trie by adding individual names."""
        # Mapping from the path to the leaf trie to the common name elements.
        # e.g. for items
        #
        #       `listing__country`,
        #       `listing__country`,
        #       `listing__country__region`
        #
        # This should contain:
        # {
        #     (`listing`,): {
        #         "country",
        #     },
        # }
        trie_path_to_common_names: dict[IndexedDunderName, set[str]] = defaultdict(set)
        for indexed_dunder_name, descriptor in items:
            name_element_count = len(indexed_dunder_name)
            if name_element_count == 0:
                raise MetricFlowInternalError(
                    LazyFormat(
                        "There must at least be one name element.",
                        name_element_list=indexed_dunder_name,
                    )
                )
            leaf_trie_path = indexed_dunder_name[:-1]

            # Traverse the trie tree to find the leaf trie.
            current_trie = self
            for name_element_index in range(name_element_count):
                current_name_element = indexed_dunder_name[name_element_index]

                # At the leaf.
                if name_element_index == name_element_count - 1:
                    name_element_to_descriptor = current_trie.name_element_to_descriptor_dict
                    current_descriptor = name_element_to_descriptor.get(current_name_element)

                    # If there are multiple names given, they are considered ambiguous.
                    # Ambiguous names should be removed.
                    if current_name_element in trie_path_to_common_names[leaf_trie_path]:
                        continue

                    if current_descriptor is None:
                        name_element_to_descriptor[current_name_element] = descriptor
                        continue

                    del name_element_to_descriptor[current_name_element]
                    trie_path_to_common_names[leaf_trie_path].add(current_name_element)
                    continue

                next_trie = current_trie.name_element_to_next_trie_dict[current_name_element]
                current_trie.name_element_to_next_trie_dict[current_name_element] = next_trie

                current_trie = next_trie

    @override
    def name_items(self, max_length: Optional[int] = None) -> Sequence[tuple[IndexedDunderName, DunderNameDescriptor]]:
        item_collector: list[tuple[IndexedDunderName, DunderNameDescriptor]] = []
        self._collect_name_items(
            name_element_prefix=(), item_collector=item_collector, current_element_index=0, max_element_count=max_length
        )
        return item_collector

    def _collect_name_items(
        self,
        name_element_prefix: IndexedDunderName,
        item_collector: list[tuple[IndexedDunderName, DunderNameDescriptor]],
        current_element_index: int,
        max_element_count: Optional[int],
    ) -> None:
        """Recursive helper method to add names represented in the trie to `item_collector`.

        If `max_element_count` is set, names with more elements than the given value are not included.
        """
        if max_element_count is not None and current_element_index + 1 > max_element_count:
            return

        item_collector.extend(
            (name_element_prefix + (name_element,), descriptor)
            for name_element, descriptor in self.name_element_to_descriptor_dict.items()
        )

        for name_element, next_node in self.name_element_to_next_trie_dict.items():
            next_node._collect_name_items(
                name_element_prefix + (name_element,), item_collector, current_element_index + 1, max_element_count
            )



from __future__ import annotations

from abc import ABC
from functools import cached_property

from typing_extensions import override

from metricflow_semantics.experimental.mf_graph.comparable import Comparable, ComparisonKey


class MetricFlowGraphLabel(Comparable, ABC):
    """Base class for objects that can be used to lookup nodes / edges in a graph."""

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return ()



from __future__ import annotations

import logging
import threading
from typing import ContextManager, Optional, Type

from metricflow_semantics.collection_helpers.mf_type_aliases import ExceptionTracebackAnyType
from metricflow_semantics.experimental.mf_graph.path_finding.pathfinder import MetricFlowPathfinder
from metricflow_semantics.experimental.mf_graph.path_finding.traversal_profile import (
    GraphTraversalProfile,
    MutableGraphTraversalProfile,
)

logger = logging.getLogger(__name__)


class TraversalProfileDiffer(ContextManager["TraversalProfileDiffer"]):
    """A context manager to help log pathfinder performance during development.

    This and associated classes are a WIP and may be removed.
    """

    def __init__(self, path_finder: MetricFlowPathfinder) -> None:  # noqa: D107
        self._path_finder = path_finder
        self._local_state = _TraversalProfileDifferLocalState()

    def __enter__(self) -> TraversalProfileDiffer:  # noqa: D105
        self._local_state.start_profile = self._path_finder.traversal_profile_snapshot
        return self

    def __exit__(  # noqa: D105
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[ExceptionTracebackAnyType],
    ) -> None:
        self._local_state.profile_delta = self._path_finder.traversal_profile_snapshot.difference(
            self._local_state.start_profile
        )

    @property
    def profile_delta(self) -> GraphTraversalProfile:  # noqa: D102
        return self._local_state.profile_delta


class _TraversalProfileDifferLocalState(threading.local):
    def __init__(self) -> None:  # noqa: D107
        self.start_profile: GraphTraversalProfile = MutableGraphTraversalProfile()
        self.profile_delta: GraphTraversalProfile = MutableGraphTraversalProfile()



from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Optional

if typing.TYPE_CHECKING:
    from metricflow_semantics.mf_logging.pretty_formatter import PrettyFormatContext


class MetricFlowPrettyFormattable(ABC):
    """Changes behavior for pretty-formatting using `MetricFlowPrettyFormatter`.

    This interface is pending updates to allow for additional configuration and structured return types.
    """

    @abstractmethod
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        """Return the pretty-formatted version of this object, or `None` if the default approach should be used."""
        raise NotImplementedError



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilterIntersection
from dbt_semantic_interfaces.protocols import WhereFilterIntersection

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.filters.time_constraint import TimeRangeConstraint
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_lookup import (
    FilterSpecResolutionLookUp,
)
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.group_by_metric_spec import GroupByMetricSpec
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.linkable_spec_set import LinkableSpecSet
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.specs.order_by_spec import OrderBySpec
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec


@dataclass(frozen=True)
class MetricFlowQuerySpec(SerializableDataclass):
    """Specs needed for running a query."""

    metric_specs: Tuple[MetricSpec, ...] = ()
    dimension_specs: Tuple[DimensionSpec, ...] = ()
    entity_specs: Tuple[EntitySpec, ...] = ()
    time_dimension_specs: Tuple[TimeDimensionSpec, ...] = ()
    group_by_metric_specs: Tuple[GroupByMetricSpec, ...] = ()
    order_by_specs: Tuple[OrderBySpec, ...] = ()
    time_range_constraint: Optional[TimeRangeConstraint] = None
    limit: Optional[int] = None
    filter_intersection: WhereFilterIntersection = field(
        default_factory=lambda: PydanticWhereFilterIntersection(where_filters=[])
    )
    filter_spec_resolution_lookup: FilterSpecResolutionLookUp = FilterSpecResolutionLookUp.empty_instance()
    min_max_only: bool = False
    apply_group_by: bool = True

    # Use the following to order the sequence of columns in the output. If a spec is not present in the list,
    # it will appear after the columns for specs that are in this field. Note that in the current implementation,
    # the ordering only applies within a group of specs of the same type. i.e. all group-by-item columns will still be
    # listed before metric columns.
    spec_output_order: AnyLengthTuple[InstanceSpec] = ()

    @property
    def linkable_specs(self) -> LinkableSpecSet:  # noqa: D102
        return LinkableSpecSet(
            dimension_specs=self.dimension_specs,
            time_dimension_specs=self.time_dimension_specs,
            entity_specs=self.entity_specs,
            group_by_metric_specs=self.group_by_metric_specs,
        )

    def with_time_range_constraint(self, time_range_constraint: Optional[TimeRangeConstraint]) -> MetricFlowQuerySpec:
        """Return a query spec that's the same as self but with a different time_range_constraint."""
        return MetricFlowQuerySpec(
            metric_specs=self.metric_specs,
            dimension_specs=self.dimension_specs,
            entity_specs=self.entity_specs,
            time_dimension_specs=self.time_dimension_specs,
            group_by_metric_specs=self.group_by_metric_specs,
            order_by_specs=self.order_by_specs,
            time_range_constraint=time_range_constraint,
            limit=self.limit,
            filter_intersection=self.filter_intersection,
            filter_spec_resolution_lookup=self.filter_spec_resolution_lookup,
        )

    def without_aliases(self) -> MetricFlowQuerySpec:
        """Return a query spec that's the same as self but with all aliases removed."""
        return MetricFlowQuerySpec(
            metric_specs=tuple(metric_spec.with_alias(None) for metric_spec in self.metric_specs),
            dimension_specs=tuple(dimension_spec.with_alias(None) for dimension_spec in self.dimension_specs),
            entity_specs=tuple(entity_spec.with_alias(None) for entity_spec in self.entity_specs),
            time_dimension_specs=tuple(
                time_dimension_spec.with_alias(None) for time_dimension_spec in self.time_dimension_specs
            ),
            group_by_metric_specs=tuple(
                group_by_metric_spec.with_alias(None) for group_by_metric_spec in self.group_by_metric_specs
            ),
            order_by_specs=tuple(order_by_spec.with_alias(None) for order_by_spec in self.order_by_specs),
            time_range_constraint=self.time_range_constraint,
            limit=self.limit,
            filter_intersection=self.filter_intersection,
            filter_spec_resolution_lookup=self.filter_spec_resolution_lookup,
            min_max_only=self.min_max_only,
            apply_group_by=self.apply_group_by,
        )



from __future__ import annotations

import logging

from typing_extensions import override

from metricflow_semantics.experimental.dsi.join_lookup import SemanticModelJoinLookup
from metricflow_semantics.experimental.dsi.manifest_object_lookup import ManifestObjectLookup
from metricflow_semantics.experimental.dsi.model_object_lookup import (
    ModelObjectLookup,
)
from metricflow_semantics.experimental.semantic_graph.builder.subgraph_generator import (
    SemanticSubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.edges.sg_edges import (
    EntityRelationshipEdge,
    JoinToModelEdge,
)
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.experimental.semantic_graph.nodes.entity_nodes import (
    ConfiguredEntityNode,
    JoinedModelNode,
    LocalModelNode,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraphEdge,
)

logger = logging.getLogger(__name__)


class EntityJoinSubgraphGenerator(SemanticSubgraphGenerator):
    """Generator for the subgraph that represents the joins that are possible between semantic models.

    Following the current query interface, joins between semantic models are modeled as a path from a model node
    to a configured-entity node, and then to another model node.

    Also see `LocalModelNode` and `JoinedModelNode` for context on why two different nodes are required
    for each semantic model.
    """

    @override
    def __init__(self, manifest_object_lookup: ManifestObjectLookup) -> None:
        super().__init__(manifest_object_lookup)
        self._verbose_debug_logs = True
        self._join_lookup = SemanticModelJoinLookup(manifest_object_lookup)

    @override
    def add_edges_for_manifest(self, edge_list: list[SemanticGraphEdge]) -> None:
        for lookup in self._manifest_object_lookup.model_object_lookups:
            self._add_edges_for_model(lookup, edge_list)

    def _add_edges_for_model(self, lookup: ModelObjectLookup, edge_list: list[SemanticGraphEdge]) -> None:
        left_model_id = SemanticModelId.get_instance(lookup.semantic_model.name)
        left_joined_model_node = JoinedModelNode.get_instance(left_model_id)
        left_local_model_node = LocalModelNode.get_instance(left_model_id)

        left_model = lookup.semantic_model

        # Add edges from the left model to the configured entities that can be reached by joining other semantic
        # models.
        #
        # e.g.
        #
        #   JoinedModel(bookings_source) -> ConfiguredEntity(listings_source.listings)
        #
        for right_model_id, join_descriptors in self._join_lookup.get_join_model_on_right_descriptors(
            left_model_id=left_model_id
        ).items():
            # Joining a semantic model to itself is currently not allowed.
            if right_model_id == left_model_id:
                continue

            # For all possible entities in the semantic model on the right.
            for join_descriptor in join_descriptors:
                right_entity_node = ConfiguredEntityNode.get_instance(
                    entity_name=join_descriptor.entity_name,
                    model_id=right_model_id,
                )
                # Add an edge from the joined-model node to the configured-entity node.
                edge_list.append(
                    JoinToModelEdge.create(
                        tail_node=left_joined_model_node,
                        head_node=right_entity_node,
                        right_model_id=right_model_id,
                    )
                )
                # Add an edge from the local-model node to the configured-entity node.
                edge_list.append(
                    JoinToModelEdge.create(
                        tail_node=left_local_model_node,
                        head_node=right_entity_node,
                        right_model_id=right_model_id,
                    )
                )

        # * The above loop created outgoing edges from the model node to the configured entity nodes.
        # * When this method is run with other models, it will do the same.
        # * From the configured-entity node, we need to add edges from the configured entity node to the model node so
        # * that there is a path from one model node to another.
        #
        # e.g.
        #   The previous loop added:
        #
        #       JoinedModel(bookings_source) -> ConfiguredEntity(listings_source.listings)
        #
        #   The following loop adds:
        #
        #       ConfiguredEntity(listings_source.listings) -> JoinedModel(listings_source)
        #
        # Combining those edges, we now have a path between the model nodes:
        #
        #   JoinedModel(bookings_source) -> ConfiguredEntity(listings_source.listings) -> JoinedModel(listings_source)
        #
        # In addition, we need to model the ability to query dimensions with an entity link, even if there's no actual
        # join. e.g. from `listings_source`, the `country_latest` dimension is accessed with `listing__country_latest`
        # but there is no join involved.
        #
        # This behavior can be modeled by adding the following edges.
        #
        #   LocalModel(listings_source) -> ConfiguredEntity(listing_source) -> JoinedModel(listings_source)
        valid_join_to_entity_types = self._join_lookup.valid_join_to_entity_types
        for entity in left_model.entities:
            if entity.type in valid_join_to_entity_types:
                entity_node = ConfiguredEntityNode.get_instance(
                    entity_name=entity.name,
                    model_id=left_model_id,
                )
                edge_list.append(
                    EntityRelationshipEdge.create(
                        tail_node=entity_node,
                        head_node=left_joined_model_node,
                    )
                )

                edge_list.append(
                    EntityRelationshipEdge.create(
                        tail_node=left_local_model_node,
                        head_node=entity_node,
                    )
                )

        # For semantic models with dimensions but no primary key, the `primary_entity` field is specified
        # as a virtual primary entity. In that case, the entity does not show up as an entity in
        # SemanticModel.entities, so add relevant edges.
        primary_entity_field_name = lookup.semantic_model.primary_entity
        if primary_entity_field_name is not None:
            primary_entity_node = ConfiguredEntityNode.get_instance(
                entity_name=primary_entity_field_name,
                model_id=left_model_id,
            )
            edge_list.append(
                EntityRelationshipEdge.create(
                    tail_node=left_local_model_node,
                    head_node=primary_entity_node,
                )
            )
            edge_list.append(
                EntityRelationshipEdge.create(
                    tail_node=primary_entity_node,
                    head_node=left_joined_model_node,
                )
            )
            # Since there's no join key, `primary_entity_node` is not accessible via a join from another model.
            # Consequently, no other edges are needed.



from __future__ import annotations

from typing import Any, Protocol


class SupportsLessThan(Protocol):
    """Protocol describing an object that supports `<`.

    This should be replaced with an already-available implementation.
    """

    def __lt__(self, other: ComparisonOtherType) -> bool:
        """Standard Python `<` comparison."""
        ...


# Type used to annotate the `other` argument in standard comparison methods like `__lt__`.
# Helpful to reduce the need to put `# type: ignore` in many places.
ComparisonOtherType = Any  # type: ignore



from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple

from typing_extensions import override

from metricflow_semantics.helpers.string_helpers import mf_indent
from metricflow_semantics.query.group_by_item.path_prefixable import PathPrefixable
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import GroupByItemResolutionNode


@dataclass(frozen=True)
class MetricFlowQueryResolutionPath(PathPrefixable):
    """Represents a path in the group-by-item resolution DAG."""

    resolution_path_nodes: Tuple[GroupByItemResolutionNode, ...]

    @staticmethod
    def empty_instance() -> MetricFlowQueryResolutionPath:  # noqa: D102
        return MetricFlowQueryResolutionPath(
            resolution_path_nodes=(),
        )

    @property
    def last_item(self) -> GroupByItemResolutionNode:
        """Return the last node in this path."""
        return self.resolution_path_nodes[-1]

    @property
    def ui_description(self) -> str:  # noqa: D102
        if len(self.resolution_path_nodes) == 0:
            return "[Empty Path]"
        # TODO: Centralize handling of error message formatting.
        max_line_length = 80
        lines = []

        # Generate text that shows where the error occurred using indents to show the nested structure.
        # e.g.
        #
        #     [Resolve Query(['bookings'])]
        #       -> [Resolve Metric('bookings')]
        #         -> [Resolve Measure('bookings')]

        for i, path_node in enumerate(self.resolution_path_nodes):
            if i == 0:
                indent_prefix = ""
            else:
                indent_prefix = mf_indent("-> ", indent_level=i)
            path_node_description = path_node.ui_description
            untruncated_line = indent_prefix + f"[Resolve {path_node_description}]"
            untruncated_line_length = len(untruncated_line)

            if untruncated_line_length > max_line_length:
                ellipsis_str = "...)"
                shorten_description_amount = untruncated_line_length - max_line_length + len(ellipsis_str)
                # Using `max()` in case of edge cases.
                shortened_description = path_node_description[
                    : max(1, len(path_node_description) - shorten_description_amount)
                ]
                lines.append(indent_prefix + f"[Resolve {shortened_description + ellipsis_str}]")
            else:
                lines.append(untruncated_line)

        return "\n".join(lines)

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> MetricFlowQueryResolutionPath:
        return MetricFlowQueryResolutionPath(
            resolution_path_nodes=path_prefix.resolution_path_nodes + self.resolution_path_nodes
        )

    @override
    def __str__(self) -> str:
        items = [self.__class__.__name__, "(", ", ".join(tuple(str(node) for node in self.resolution_path_nodes)), ")"]
        return "".join(items)

    @staticmethod
    def from_path_item(node: GroupByItemResolutionNode) -> MetricFlowQueryResolutionPath:
        """Creates a single element path with the given node."""
        return MetricFlowQueryResolutionPath(
            resolution_path_nodes=(node,),
        )



from __future__ import annotations

import datetime
import gc
import logging
import time
import timeit

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


def assert_performance_factor(
    left_setup: str,
    left_statement: str,
    right_setup: str,
    right_statement: str,
    min_performance_factor: float,
) -> None:
    """Using `timeit`, check that the right statement is faster than the left statement by the given factor.

    See `timeit` for definitions of `setup` and `statement`.
    """
    attempt_count = 3
    sleep_time = 1.0
    for i in range(attempt_count):
        try:
            # `autorange()` runs the given code multiple times until 0.2s is spent.
            gc.collect()
            iteration_count, total_time = timeit.Timer(
                timer=time.thread_time, setup=left_setup, stmt=left_statement
            ).autorange()
            average_left_code_runtime = total_time / iteration_count

            gc.collect()
            iteration_count, total_time = timeit.Timer(
                timer=time.thread_time, setup=right_setup, stmt=right_statement
            ).autorange()
            average_right_code_runtime = total_time / iteration_count

        except Exception as e:
            raise RuntimeError(
                LazyFormat(
                    "Got an exception with the executed statements.",
                    left_setup=left_setup,
                    left_statement=left_statement,
                    right_setup=right_setup,
                    right_statement=right_statement,
                )
            ) from e

        performance_factor = average_left_code_runtime / average_right_code_runtime

        logger.debug(
            LazyFormat(
                "Compared performance of left and right statements.",
                left_statement_runtime=lambda: datetime.timedelta(seconds=average_left_code_runtime),
                right_statement_runtime=lambda: datetime.timedelta(seconds=average_right_code_runtime),
                performance_factor=lambda: f"{performance_factor:.2f}",
                expected_min_performance_factor=lambda: f"{min_performance_factor:.2f}",
            )
        )

        if performance_factor > min_performance_factor:
            return

        if i == attempt_count - 1:
            raise AssertionError(
                LazyFormat(
                    "The performance factor is lower than the minimum expected.",
                    expected_min_performance_factor=min_performance_factor,
                    actual_performance_factor=lambda: f"{performance_factor:.2f}",
                    attempt_count=attempt_count,
                )
            )
        else:
            logger.debug(
                LazyFormat(
                    "Since performance is lower than expected, sleeping and retrying in case it's a fluke",
                    sleep_time=lambda: f"{sleep_time:.2f}s",
                    attempt_num=i + 1,
                    attempt_count=attempt_count,
                )
            )
            time.sleep(sleep_time)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Sequence

from dbt_semantic_interfaces.protocols.metric import MetricTimeWindow
from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class OffsetBaseGrainByCustomGrainNode(DataflowPlanNode, ABC):
    """For a given custom grain, offset its base grain by the requested number of custom grain periods.

    Used to build the time spine node when querying a metric with a custom offset window with any grains / date parts.
    This node should only satisfy base grains - custom grains will be joined later in the dataflow plan. It will also
    need to include the custom grain's base grain, even if that was not queried, because that will be required to join
    to the source node.


    If the metric is queried with ONLY the same grain as is used in the offset window, should use OffsetCustomGrainNode
    instead.
    """

    offset_window: MetricTimeWindow
    required_time_spine_specs: Sequence[TimeDimensionSpec]
    time_spine_node: DataflowPlanNode

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()

        for spec in self.required_time_spine_specs:
            if spec.has_custom_grain:
                raise ValueError(
                    LazyFormat(
                        "Found custom grain in required specs, which is not supported by OffsetBaseGrainByCustomGrainNode.",
                        required_time_spine_specs=self.required_time_spine_specs,
                    )
                )
        if self.offset_window.is_standard_granularity:
            raise ValueError(
                LazyFormat(
                    "OffsetBaseGrainByCustomGrainNode should only be used for custom grain offset windows.",
                    offset_window=self.offset_window,
                )
            )

    @staticmethod
    def create(  # noqa: D102
        time_spine_node: DataflowPlanNode,
        offset_window: MetricTimeWindow,
        required_time_spine_specs: Sequence[TimeDimensionSpec],
    ) -> OffsetBaseGrainByCustomGrainNode:
        return OffsetBaseGrainByCustomGrainNode(
            parent_nodes=(time_spine_node,),
            time_spine_node=time_spine_node,
            offset_window=offset_window,
            required_time_spine_specs=required_time_spine_specs,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_OFFSET_BY_CUSTOM_GRANULARITY_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_offset_base_grain_by_custom_grain_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Offset Base Granularity By Custom Granularity Period(s)"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (
            DisplayedProperty("offset_window", self.offset_window),
            DisplayedProperty("required_time_spine_specs", self.required_time_spine_specs),
        )

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and other_node.offset_window == self.offset_window
            and other_node.required_time_spine_specs == self.required_time_spine_specs
        )

    def with_new_parents(  # noqa: D102
        self, new_parent_nodes: Sequence[DataflowPlanNode]
    ) -> OffsetBaseGrainByCustomGrainNode:
        assert len(new_parent_nodes) == 1
        return OffsetBaseGrainByCustomGrainNode(
            parent_nodes=tuple(new_parent_nodes),
            time_spine_node=new_parent_nodes[0],
            offset_window=self.offset_window,
            required_time_spine_specs=self.required_time_spine_specs,
        )



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import Optional

from dbt_semantic_interfaces.protocols import MetricTimeWindow
from dbt_semantic_interfaces.type_enums import TimeGranularity

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.model.semantics.simple_metric_input import SimpleMetricInput
from metricflow_semantics.specs.instance_spec import InstanceSpec, InstanceSpecVisitor
from metricflow_semantics.specs.where_filter.where_filter_spec_set import WhereFilterSpecSet
from metricflow_semantics.sql.sql_join_type import SqlJoinType
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class SimpleMetricInputSpec(InstanceSpec):  # noqa: D101
    element_name: str
    fill_nulls_with: Optional[int] = None

    @property
    def dunder_name(self) -> str:  # noqa: D102
        return self.element_name

    def accept(self, visitor: InstanceSpecVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_simple_metric_input_spec(self)


@dataclass(frozen=True)
class CumulativeDescription:
    """If a simple metric is a part of a cumulative metric, this represents the associated parameters."""

    cumulative_window: Optional[MetricTimeWindow]
    cumulative_grain_to_date: Optional[TimeGranularity]


@fast_frozen_dataclass()
class SimpleMetricRecipe:
    """Describes how to build a simple metric but with modifications."""

    simple_metric_input: SimpleMetricInput

    # For the filter defined in `simple_metric_input`
    metric_filter_spec_set: WhereFilterSpecSet
    # For additional filters that might be needed (e.g. a filter defined in a derived metric or query).
    additional_filter_spec_set: WhereFilterSpecSet

    offset_window: Optional[MetricTimeWindow]
    offset_to_grain: Optional[TimeGranularity]
    cumulative_description: Optional[CumulativeDescription]
    alias: Optional[str]
    before_aggregation_time_spine_join_description: Optional[JoinToTimeSpineDescription]
    after_aggregation_time_spine_join_description: Optional[JoinToTimeSpineDescription]

    @cached_property
    def combined_filter_spec_set(self) -> WhereFilterSpecSet:  # noqa: D102
        return self.metric_filter_spec_set.merge(self.additional_filter_spec_set)


@dataclass(frozen=True)
class JoinToTimeSpineDescription:
    """Describes how a time spine join should be performed."""

    join_type: SqlJoinType
    offset_window: Optional[MetricTimeWindow]
    offset_to_grain: Optional[TimeGranularity]

    @property
    def standard_offset_window(self) -> Optional[MetricTimeWindow]:
        """Return the offset window if it uses a standard granularity."""
        if self.offset_window and self.offset_window.is_standard_granularity:
            return self.offset_window
        return None

    @property
    def custom_offset_window(self) -> Optional[MetricTimeWindow]:
        """Return the offset window if it uses a custom granularity."""
        if self.offset_window and not self.offset_window.is_standard_granularity:
            return self.offset_window
        return None

    @property
    def uses_offset(self) -> bool:
        """Return True if the simple-metric input uses an offset."""
        return self.offset_window is not None or self.offset_to_grain is not None



from __future__ import annotations

from typing import Any, Hashable, TypeVar

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)

T1 = TypeVar("T1")
T2 = TypeVar("T2")

T1_co = TypeVar("T1_co", covariant=True)
T2_co = TypeVar("T2_co", covariant=True)


HashableT = TypeVar("HashableT", bound=Hashable)
HashableT_co = TypeVar("HashableT_co", bound=Hashable, covariant=True)

# A pair of objects.
Pair = tuple[T1, T2]

# A tuple of any length. Faster to auto-complete than typing out `, ...`
AnyLengthTuple = tuple[T1, ...]

# A tuple of items in a mapping. Each item is a tuple of the form `(key, value)`
# These are useful for type-annotating immutable dataclasses where a mapping is stored as a tuple.
KeyT = TypeVar("KeyT", bound=Hashable)
ValueT = TypeVar("ValueT")
MappingItem = tuple[KeyT, ValueT]
MappingItemsTuple = tuple[tuple[KeyT, ValueT], ...]

ExceptionTracebackAnyType = Any  # type: ignore



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from typing_extensions import override

from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable
from metricflow_semantics.mf_logging.pretty_formatter import PrettyFormatContext

logger = logging.getLogger(__name__)


class GraphTraversalProfile(MetricFlowPrettyFormattable, ABC):
    """A read-only interface for a profile containing counters related to graph traversal.

    Used for debugging / logs.
    """

    @property
    @abstractmethod
    def visited_nodes_count(self) -> int:  # noqa: D102
        raise NotImplementedError()

    @property
    @abstractmethod
    def examined_edges_count(self) -> int:  # noqa: D102
        raise NotImplementedError()

    @property
    @abstractmethod
    def generated_paths_count(self) -> int:  # noqa: D102
        raise NotImplementedError()

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        return format_context.formatter.pretty_format(
            {
                "node_visit_count": self.visited_nodes_count,
                "edge_examination_count": self.examined_edges_count,
                "path_creation_count": self.generated_paths_count,
            }
        )

    @abstractmethod
    def difference(self, other: GraphTraversalProfile) -> GraphTraversalProfile:  # noqa: D102
        raise NotImplementedError


@dataclass
class MutableGraphTraversalProfile(GraphTraversalProfile):
    """A mutable version of `GraphTraversalProfile` with counters that can be incremented."""

    _visited_nodes_count: int = 0
    _examined_edges_count: int = 0
    _generated_paths_count: int = 0

    def increment_node_visit_count(self, visit_count: int = 1) -> None:  # noqa: D102
        self._visited_nodes_count += visit_count

    def increment_edge_examined_count(self, examined_count: int = 1) -> None:  # noqa: D102
        self._examined_edges_count += examined_count

    def increment_generated_paths_count(self, generated_count: int = 1) -> None:  # noqa: D102
        self._generated_paths_count += generated_count

    @property
    def visited_nodes_count(self) -> int:  # noqa: D102
        return self._visited_nodes_count

    @property
    def examined_edges_count(self) -> int:  # noqa: D102
        return self._visited_nodes_count

    @property
    def generated_paths_count(self) -> int:  # noqa: D102
        return self._generated_paths_count

    @override
    def difference(self, other: GraphTraversalProfile) -> GraphTraversalProfile:  # noqa: D102
        return MutableGraphTraversalProfile(
            _visited_nodes_count=self._visited_nodes_count - other.visited_nodes_count,
            _examined_edges_count=self._examined_edges_count - other.examined_edges_count,
            _generated_paths_count=self._generated_paths_count - other.generated_paths_count,
        )

    def copy(self) -> MutableGraphTraversalProfile:  # noqa: D102
        return MutableGraphTraversalProfile(
            _visited_nodes_count=self._visited_nodes_count,
            _examined_edges_count=self._examined_edges_count,
            _generated_paths_count=self._generated_paths_count,
        )



from __future__ import annotations

import logging
from collections.abc import Mapping
from functools import cached_property

from dbt_semantic_interfaces.type_enums import DatePart, TimeGranularity
from typing_extensions import override

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.dsi.manifest_object_lookup import ManifestObjectLookup
from metricflow_semantics.experimental.dsi.model_object_lookup import (
    ModelObjectLookup,
)
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe_step import (
    AttributeRecipeStep,
)
from metricflow_semantics.experimental.semantic_graph.builder.subgraph_generator import (
    SemanticSubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.edges.sg_edges import EntityRelationshipEdge
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.experimental.semantic_graph.nodes.entity_nodes import (
    JoinedModelNode,
    MetricTimeNode,
    TimeDimensionNode,
    TimeNode,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraphEdge,
)

logger = logging.getLogger(__name__)


class TimeDimensionSubgraphGenerator(SemanticSubgraphGenerator):
    """Generator to build edges for time-dimension entities.

    A time dimension in a semantic model maps to a time-dimension entity node. All time-dimension entity nodes relate
    to the time-entity node.
    """

    @override
    def __init__(self, manifest_object_lookup: ManifestObjectLookup) -> None:
        super().__init__(manifest_object_lookup)
        self._time_entity_node = TimeNode.get_instance()

    @override
    def add_edges_for_manifest(self, edge_list: list[SemanticGraphEdge]) -> None:
        for lookup in self._manifest_object_lookup.model_object_lookups:
            self._add_edges_for_model(lookup, edge_list)

        edge_list.append(
            EntityRelationshipEdge.create(tail_node=MetricTimeNode.get_instance(), head_node=self._time_entity_node)
        )

    def _add_edges_for_model(self, lookup: ModelObjectLookup, edge_list: list[SemanticGraphEdge]) -> None:
        model_id = SemanticModelId.get_instance(model_name=lookup.semantic_model.name)
        semantic_model_node = JoinedModelNode.get_instance(model_id)

        for time_dimension_name, time_grain in lookup.time_dimension_name_to_grain.items():
            time_dimension_node = TimeDimensionNode.get_instance(time_dimension_name)
            edge_list.append(
                EntityRelationshipEdge.create(
                    tail_node=semantic_model_node,
                    head_node=time_dimension_node,
                    recipe_update=AttributeRecipeStep(
                        set_source_time_grain=time_grain,
                    ),
                )
            )
            edge_list.append(
                EntityRelationshipEdge.create(
                    tail_node=time_dimension_node,
                    head_node=self._time_entity_node,
                )
            )

    @cached_property
    def _time_grain_to_queryable_time_grains(self) -> Mapping[TimeGranularity, AnyLengthTuple[TimeGranularity]]:
        return {
            min_referenced_time_grain: tuple(
                time_grain
                for time_grain in TimeGranularity
                if time_grain.to_int() >= min_referenced_time_grain.to_int()
            )
            for min_referenced_time_grain in TimeGranularity
        }

    @cached_property
    def _time_grain_to_applicable_date_parts(self) -> Mapping[TimeGranularity, AnyLengthTuple[DatePart]]:
        return {
            time_grain: tuple(date_part for date_part in DatePart if date_part.to_int() >= time_grain.to_int())
            for time_grain in TimeGranularity
        }



from __future__ import annotations

import logging

from typing_extensions import override

from metricflow_semantics.experimental.dsi.simple_metric_model_object_lookup import SimpleMetricModelObjectLookup
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe_step import (
    AttributeRecipeStep,
)
from metricflow_semantics.experimental.semantic_graph.builder.subgraph_generator import (
    SemanticSubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.edges.sg_edges import EntityRelationshipEdge
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.experimental.semantic_graph.nodes.entity_nodes import (
    LocalModelNode,
    MetricTimeNode,
    SimpleMetricNode,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraphEdge,
)

logger = logging.getLogger(__name__)


class SimpleMetricSubgraphGenerator(SemanticSubgraphGenerator):
    """Generate the subgraph that relates simple metrics to other entities.

    For each simple metric, this adds the following edges:

    * SimpleMetricNode -> LocalModelNode (corresponding to the semantic model where the simple metric is defined)
    * SimpleMetricNode -> MetricTimeNode (edge describes aggregation time grain)
    """

    @override
    def add_edges_for_manifest(self, edge_list: list[SemanticGraphEdge]) -> None:
        for lookup in self._manifest_object_lookup.simple_metric_model_lookups:
            self._add_edges_for_simple_metric_model(lookup, edge_list)

    def _add_edges_for_simple_metric_model(
        self, lookup: SimpleMetricModelObjectLookup, edge_list: list[SemanticGraphEdge]
    ) -> None:
        model_id = SemanticModelId.get_instance(model_name=lookup.semantic_model.name)
        local_model_node = LocalModelNode.get_instance(model_id)

        metric_time_node = MetricTimeNode.get_instance()
        # For each unique aggregation time grain:
        for (
            aggregation_configuration,
            simple_metric_inputs,
        ) in lookup.aggregation_configuration_to_simple_metric_inputs.items():
            # For all simple-metric inputs:
            for simple_metric_input in simple_metric_inputs:
                simple_metric_name = simple_metric_input.name

                simple_metric_node = SimpleMetricNode.get_instance(simple_metric_name)
                # Add an edge from the simple-metric input node to the metric-time node.
                edge_list.append(
                    EntityRelationshipEdge.create(
                        tail_node=simple_metric_node,
                        head_node=metric_time_node,
                        recipe_update=AttributeRecipeStep(
                            set_source_time_grain=aggregation_configuration.time_grain,
                            add_model_join=model_id,
                        ),
                    )
                )
                # Add an edge from the simple-metric input node to the model node.
                edge_list.append(
                    EntityRelationshipEdge.create(
                        tail_node=simple_metric_node,
                        head_node=local_model_node,
                    )
                )



from __future__ import annotations

import logging
from collections.abc import Mapping
from functools import cached_property
from typing import Optional, Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.specs.simple_metric_input_spec import SimpleMetricInputSpec, SimpleMetricRecipe
from typing_extensions import override

logger = logging.getLogger(__name__)


@fast_frozen_dataclass()
class InstanceAliasMapping(Mergeable, SerializableDataclass):
    """Stores the mapping for how instances should be aliased (i.e. given a different column name)."""

    _element_name_and_alias_items: Tuple[Tuple[str, str], ...]

    @staticmethod
    def create(element_name_to_alias: Optional[Mapping[str, str]] = None) -> InstanceAliasMapping:  # noqa: D102
        if element_name_to_alias is None:
            return InstanceAliasMapping(_element_name_and_alias_items=())

        return InstanceAliasMapping(
            _element_name_and_alias_items=tuple(
                (element_name, alias) for element_name, alias in element_name_to_alias.items()
            )
        )

    @staticmethod
    def create_from_simple_metric_recipe(  # noqa: D102
        simple_metric_recipe: SimpleMetricRecipe,
    ) -> InstanceAliasMapping:
        alias = simple_metric_recipe.alias
        if alias is None:
            return InstanceAliasMapping(_element_name_and_alias_items=())

        return InstanceAliasMapping(
            _element_name_and_alias_items=((simple_metric_recipe.simple_metric_input.name, alias),)
        )

    @cached_property
    def element_name_to_alias(self) -> Mapping[str, str]:  # noqa: D102
        return {element_name: alias for element_name, alias in self._element_name_and_alias_items}

    def aliased_spec(self, spec: SimpleMetricInputSpec) -> Optional[SimpleMetricInputSpec]:
        """If this mapping contains an alias for the given spec, return the spec with the alias."""
        alias = self.element_name_to_alias.get(spec.element_name)
        if alias is None:
            return None
        return SimpleMetricInputSpec(
            element_name=alias,
            fill_nulls_with=spec.fill_nulls_with,
        )

    @override
    def merge(self, other: InstanceAliasMapping) -> InstanceAliasMapping:
        if self.has_conflict(other):
            raise RuntimeError(
                LazyFormat(
                    "Can't merge alias mappings with conflicting items",
                    self=self,
                    other=other,
                )
            )
        return InstanceAliasMapping.create({**self.element_name_to_alias, **other.element_name_to_alias})

    @classmethod
    @override
    def empty_instance(cls) -> InstanceAliasMapping:
        return InstanceAliasMapping(_element_name_and_alias_items=())

    def has_conflict(self, other: InstanceAliasMapping) -> bool:  # noqa: D102
        other_element_name_to_alias = other.element_name_to_alias
        other_aliases = set(other.element_name_to_alias.values())
        for element_name, alias in self.element_name_to_alias.items():
            other_alias = other_element_name_to_alias.get(element_name)
            if other_alias is not None:
                if alias != other_alias:
                    return True
            elif alias in other_aliases:
                return True
        return False


@fast_frozen_dataclass()
class NullFillValueMapping(Mergeable, SerializableDataclass):
    """Stores the mapping for which instances should have null values set to a configured value."""

    _element_name_and_null_fill_value_items: Tuple[Tuple[str, int], ...]

    @staticmethod
    def create(  # noqa: D102
        element_name_to_null_fill_value: Optional[Mapping[str, int]] = None
    ) -> NullFillValueMapping:  # noqa: D102
        if element_name_to_null_fill_value is None:
            return NullFillValueMapping(_element_name_and_null_fill_value_items=())
        return NullFillValueMapping(
            _element_name_and_null_fill_value_items=tuple(
                (element_name, null_fill_value)
                for element_name, null_fill_value in element_name_to_null_fill_value.items()
            )
        )

    @staticmethod
    def create_from_simple_metric_recipe(  # noqa: D102
        simple_metric_recipe: SimpleMetricRecipe,
    ) -> NullFillValueMapping:
        null_fill_value = simple_metric_recipe.simple_metric_input.fill_nulls_with
        if null_fill_value is None:
            return NullFillValueMapping(_element_name_and_null_fill_value_items=())

        return NullFillValueMapping(
            _element_name_and_null_fill_value_items=((simple_metric_recipe.simple_metric_input.name, null_fill_value),)
        )

    @cached_property
    def element_name_to_null_fill_value(self) -> Mapping[str, int]:  # noqa: D102
        result = {}
        for element_name, alias in self._element_name_and_null_fill_value_items:
            result[element_name] = alias
        return result

    def null_fill_value_spec(self, spec: SimpleMetricInputSpec) -> Optional[SimpleMetricInputSpec]:
        """If this mapping contains a null fill value for the given spec, return the spec with the value set."""
        null_fill_value = self.element_name_to_null_fill_value.get(spec.element_name)
        if null_fill_value is None:
            return None
        return SimpleMetricInputSpec(
            element_name=spec.element_name,
            fill_nulls_with=null_fill_value,
        )

    @override
    def merge(self, other: NullFillValueMapping) -> NullFillValueMapping:
        element_name_to_null_fill_value: dict[str, int] = {**self.element_name_to_null_fill_value}
        for element_name, other_null_fill_value in other._element_name_and_null_fill_value_items:
            current_null_fill_value = element_name_to_null_fill_value.get(element_name)
            if current_null_fill_value is not None and current_null_fill_value != other_null_fill_value:
                raise RuntimeError(
                    LazyFormat(
                        "Can't merge fill value mappings with conflicting items.",
                        element_name=element_name,
                        current_null_fill_value=current_null_fill_value,
                        other_null_fill_value=other_null_fill_value,
                    )
                )
            element_name_to_null_fill_value[element_name] = other_null_fill_value
        return NullFillValueMapping.create(element_name_to_null_fill_value)

    @classmethod
    @override
    def empty_instance(cls) -> NullFillValueMapping:
        return NullFillValueMapping(_element_name_and_null_fill_value_items=())



from __future__ import annotations

import itertools
import logging
from functools import cached_property
from typing import Mapping, Optional, Set

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.type_enums import DatePart, TimeGranularity
from typing_extensions import override

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.mf_graph.path_finding.weight_function import WeightFunction
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe import AttributeRecipe
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe_step import (
    AttributeRecipeStep,
)
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.recipe_writer_path import (
    AttributeRecipeWriterPath,
)
from metricflow_semantics.experimental.semantic_graph.nodes.node_labels import (
    GroupByAttributeLabel,
    TimeClusterLabel,
    TimeDimensionLabel,
)
from metricflow_semantics.experimental.semantic_graph.sg_exceptions import SemanticGraphTraversalError
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import SemanticGraphEdge, SemanticGraphNode
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty
from metricflow_semantics.model.semantics.element_filter import GroupByItemSetFilter
from metricflow_semantics.model.semantics.linkable_element import LinkableElementType
from metricflow_semantics.model.semantics.semantic_model_join_evaluator import MAX_JOIN_HOPS

logger = logging.getLogger(__name__)


class AttributeRecipeWriterWeightFunction(
    WeightFunction[SemanticGraphNode, SemanticGraphEdge, AttributeRecipeWriterPath]
):
    """The weight function to use with `AttributeRecipeWriterPath`.

    The weight function maps a path in the semantic graph to the number of entity links in the dunder name associated
    with the path / attribute.

    In addition, this weight function models the characteristics of the query interface by returning `None` in certain
    cases. `None` signals to the pathfinder that the edge is blocked. This weight function returns `None` to prohibit:

    * Repeated name elements in a dunder name (e.g. `listing__listing`).
    * Repeated semantic models in a join.
    * Querying a time dimension at a time grain that's smaller than the one configured in the semantic model.

    In addition, this can be initialized with a `GroupByItemSetFilter` to handle filtering of results as required by
    the MF engine API.

    As the methods in this class are called repeatedly in a relatively tight loop during resolution, the performance of
    this class is critical. Hence, there are a number of early-exit checks.
    """

    def __init__(  # noqa: D107
        self, element_filter: Optional[GroupByItemSetFilter] = None, max_path_model_count: Optional[int] = None
    ) -> None:
        """Initializer.

        Args:
            element_filter: If specified, limits paths to align with the given filter.
            max_path_model_count: To handle some edge cases, this can be specified to limit the number of semantic
            models in the path.
        """
        self._element_filter = element_filter
        self._group_by_attribute_label = GroupByAttributeLabel.get_instance()
        self._time_dimension_label = TimeDimensionLabel.get_instance()
        self._time_cluster_label = TimeClusterLabel.get_instance()
        self._max_path_model_count = max_path_model_count

        self._verbose_debug_logs = False

    @override
    def incremental_weight(
        self,
        path_to_node: AttributeRecipeWriterPath,
        next_edge: SemanticGraphEdge,
    ) -> Optional[int]:
        next_node = next_edge.head_node
        current_recipe = path_to_node.latest_recipe
        # First run checks that can be done without the next recipe (the profiler showed non-insignificant time spent
        # generating recipes).
        next_edge_step = next_edge.recipe_step_to_append
        next_node_step = next_node.recipe_step_to_append

        element_filter = self._element_filter

        # We do not allow repeated element names in the dundered name (e.g. `listing__listing`).
        if AttributeRecipeWriterWeightFunction.repeated_dunder_name_elements(
            current_recipe, next_edge_step, next_node_step
        ):
            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Blocking edge due to a repeated name element.",
                        next_edge_step=next_edge_step,
                        next_node_step=next_node_step,
                    )
                )
            return None

        # Don't allow joining a semantic model multiple times.
        if AttributeRecipeWriterWeightFunction.repeated_model_join(current_recipe, next_edge_step, next_node_step):
            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Blocking edge due to a repeated model.",
                        models_in_join=current_recipe.joined_model_ids,
                        next_edge_step=next_edge_step,
                        next_node_step=next_node_step,
                    )
                )
            return None

        # A quick check to see if the filter will block the edge.
        if element_filter and self._filter_denies_edge(element_filter, next_node, (next_edge_step, next_node_step)):
            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Blocking edge due to the element filter.",
                        next_edge=next_edge,
                        next_edge_step=next_edge_step,
                        next_node_step=next_node_step,
                        element_filter=self._element_filter,
                    )
                )
            return None

        weight_added_by_taking_edge = 0
        if next_edge.recipe_step_to_append.add_entity_link is not None:
            weight_added_by_taking_edge += 1
        if next_edge.head_node.recipe_step_to_append.add_entity_link is not None:
            weight_added_by_taking_edge += 1

        if len(current_recipe.entity_link_names) + weight_added_by_taking_edge > MAX_JOIN_HOPS:
            return None

        # Check if this needs to limit joins.
        if self._max_path_model_count is not None:
            current_path_model_count = len(current_recipe.joined_model_ids)
            if next_edge_step.add_model_join:
                current_path_model_count += 1
            if next_node_step.add_model_join:
                current_path_model_count += 1
            if current_path_model_count > self._max_path_model_count:
                return None

        # If the current path is not yet at an attribute node, we can't run the checks below so return early.
        next_node_is_attribute = self._group_by_attribute_label in next_node.labels
        if not next_node_is_attribute:
            return weight_added_by_taking_edge

        next_edge_update = next_edge.recipe_step_to_append
        next_node_update = next_edge.head_node.recipe_step_to_append
        next_recipe = current_recipe.append_step(next_edge_update).append_step(next_node_update)

        # Require entity links for dimensions / time dimensions, except for metric time
        if self._invalid_entity_links(next_recipe):
            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Blocking edge due to entity links being outside of range.",
                        element_type=next_recipe.element_type,
                        entity_link_names=next_recipe.entity_link_names,
                        models_in_join=next_recipe.joined_model_ids,
                    )
                )
            return None

        if self._source_time_grain_mismatch(next_recipe):
            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Blocking edge due to source time-grain mismatch.",
                        next_edge=next_edge,
                        next_recipe=next_recipe,
                    )
                )
            return None

        if element_filter and self._filter_denies_recipe(element_filter, next_recipe):
            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Blocking edge due to the element filter.",
                        next_edge=next_edge,
                        next_recipe=next_recipe,
                        element_filter=self._element_filter,
                    )
                )
            return None

        return weight_added_by_taking_edge

    def _filter_denies_edge(
        self,
        element_filter: GroupByItemSetFilter,
        next_node: SemanticGraphNode,
        steps: AnyLengthTuple[AttributeRecipeStep],
    ) -> bool:
        any_properties_denylist = element_filter.any_properties_denylist
        if any_properties_denylist and any(
            element_property in any_properties_denylist
            for element_property in itertools.chain.from_iterable(
                step.add_properties for step in steps if step.add_properties
            )
        ):
            return True

        name_element_allow_set = element_filter.element_name_allowlist
        # Generally, element name should be checked at the attribute node, but it can be done at time dimension nodes
        # as they are an entity node that adds an element name. This can speed up traversal by pruning edges.
        if (
            name_element_allow_set
            and self._time_dimension_label in next_node.labels
            and not all(
                name_element in name_element_allow_set
                for name_element in (
                    step.add_dunder_name_element for step in steps if step.add_dunder_name_element is not None
                )
            )
        ):
            return True

        return False

    def _filter_denies_recipe(
        self,
        element_filter: GroupByItemSetFilter,
        next_recipe: AttributeRecipe,
    ) -> bool:
        """Check if the element filter denies the recipe associated with the path.

        As per call in the incremental weight function, this should only be called when the next node is an attribute
        node.
        """
        next_property_set = next_recipe.resolve_complete_properties()
        next_element_name = next_recipe.resolve_element_name()
        if next_property_set is None or next_element_name is None:
            raise SemanticGraphTraversalError(
                LazyFormat(
                    "Expected the recipe at an attribute node or a time-attribute node to have complete"
                    " properties and an element name, but it is missing at least one of them.",
                    next_property_set=next_property_set,
                    element_name=next_element_name,
                    next_recipe=next_recipe,
                )
            )

        return not element_filter.allow(next_element_name, next_property_set)

    @staticmethod
    def repeated_dunder_name_elements(
        recipe: AttributeRecipe,
        step: AttributeRecipeStep,
        other_step: AttributeRecipeStep,
    ) -> bool:
        """Return true if the recipe + steps would cause a repeated element in the dunder name."""
        next_edge_add_dunder_name_element = step.add_dunder_name_element
        next_node_add_dunder_name_element = other_step.add_dunder_name_element
        if next_edge_add_dunder_name_element is None and next_node_add_dunder_name_element is None:
            return False

        current_dunder_name_elements = recipe.dunder_name_elements_set
        return (
            next_node_add_dunder_name_element
            and next_node_add_dunder_name_element in current_dunder_name_elements
            or next_edge_add_dunder_name_element
            and next_edge_add_dunder_name_element in current_dunder_name_elements
            or next_edge_add_dunder_name_element == next_node_add_dunder_name_element
        )

    @staticmethod
    def repeated_model_join(
        recipe: AttributeRecipe,
        step: AttributeRecipeStep,
        other_step: AttributeRecipeStep,
    ) -> bool:
        """Return true if the recipe + steps would cause a repeated semantic model in the join."""
        edge_add_model_join = step.add_model_join
        node_add_model_join = other_step.add_model_join
        if edge_add_model_join is None and node_add_model_join is None:
            return False
        if edge_add_model_join == node_add_model_join:
            return True

        for model_in_join in recipe.joined_model_ids:
            if model_in_join == edge_add_model_join or model_in_join == node_add_model_join:
                return True
        return False

    def _invalid_entity_links(self, next_recipe: AttributeRecipe) -> bool:
        # Require entity links for dimensions / time dimensions, except for metric time
        next_recipe_element_type = next_recipe.element_type
        if next_recipe_element_type is None:
            return False

        # Doing a max for the join count to allow this weight function to be used for paths that haven't yet
        # included a semantic model.
        join_count = max(0, len(next_recipe.joined_model_ids) - 1)
        min_entity_link_length = 1
        if join_count == 0:
            max_entity_link_length = 1
        else:
            max_entity_link_length = join_count

        if next_recipe_element_type is LinkableElementType.ENTITY:
            min_entity_link_length = 0
        elif (
            next_recipe_element_type is LinkableElementType.TIME_DIMENSION
            or next_recipe_element_type is LinkableElementType.DIMENSION
        ):
            if GroupByItemProperty.METRIC_TIME in next_recipe.element_properties:
                min_entity_link_length = 0

        elif next_recipe_element_type is LinkableElementType.METRIC:
            pass
        else:
            assert_values_exhausted(next_recipe_element_type)

        next_entity_link_length = len(next_recipe.entity_link_names)
        return not (min_entity_link_length <= next_entity_link_length <= max_entity_link_length)

    def _source_time_grain_mismatch(self, next_recipe: AttributeRecipe) -> bool:
        source_time_grain = next_recipe.source_time_grain
        recipe_time_grain = next_recipe.recipe_time_grain
        if source_time_grain is None:
            return False

        if recipe_time_grain is not None and recipe_time_grain.base_granularity.to_int() < source_time_grain.to_int():
            return True

        date_part = next_recipe.recipe_date_part
        if date_part is None:
            return False

        return source_time_grain not in self._date_part_to_min_time_grain[date_part]

    @cached_property
    def _date_part_to_min_time_grain(self) -> Mapping[DatePart, Set[TimeGranularity]]:
        return {date_part: set(date_part.compatible_granularities) for date_part in DatePart}



from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import Iterable, TypeVar

from metricflow_semantics.collection_helpers.mf_type_aliases import (
    KeyT,
    MappingItemsTuple,
    Pair,
    ValueT,
)

logger = logging.getLogger(__name__)


T1 = TypeVar("T1")
T2 = TypeVar("T2")


def mf_mapping_to_items(mapping: Mapping[KeyT, ValueT]) -> MappingItemsTuple[KeyT, ValueT]:
    """Converts the items in a mapping to tuples."""
    return tuple((key, value) for key, value in mapping.items())


def mf_items_to_dict(items: Iterable[Pair[KeyT, ValueT]]) -> dict[KeyT, ValueT]:
    """Convert mapping items represented as tuples to a dict."""
    return {key: value for key, value in items}



from __future__ import annotations

import logging
import os

import _pytest.fixtures

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.test_helpers.snapshot_helpers import SnapshotConfiguration, snapshot_path_prefix
from metricflow_semantics.test_helpers.terminal_helpers import mf_colored_link_text

logger = logging.getLogger(__name__)


def write_svg_snapshot_for_review(
    request: _pytest.fixtures.FixtureRequest,
    snapshot_configuration: SnapshotConfiguration,
    svg_file_contents: str,
) -> None:
    """Write a snapshot of the SVG file contents to aid review.

    * In a Github PR, SVG files are displayed as images.
    * The snapshot is not compared as SVG rendering via `graphviz` is not deterministic (e.g. between platforms).
    * These snapshots should be deleted before merging. Considering adding Github actions to help with this.
    """
    file_path = snapshot_path_prefix(
        request=request,
        snapshot_configuration=snapshot_configuration,
        snapshot_group="svg",
        snapshot_id="result",
    ).with_suffix(".svg")

    if snapshot_configuration.overwrite_snapshots:
        # Create parent directory for the plan text files.
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w") as snapshot_text_file:
            snapshot_text_file.write(svg_file_contents)
        logger.debug(
            LazyFormat(
                "Wrote SVG for PR review",
                file_path=file_path,
                open_link=mf_colored_link_text(file_path.resolve().as_uri()),
                iterm_hint="Link may be opened with <Command> + <Left Click>",
            )
        )
    else:
        logger.debug(
            LazyFormat(
                "Since this session is not overwriting snapshots, leaving snapshot file alone", file_path=file_path
            )
        )
